from microdf import MicroDataFrame
import synthimpute as si
from policyengine_core.data.dataset import Dataset
from policyengine_uk.data.datasets.frs.frs import FRS
from policyengine_uk.data.datasets.spi.spi import SPI

PREDICTORS = [
    "age",
    "region",
]

IMPUTATIONS = [
    "employment_income",
    "self_employment_income",
    "savings_interest_income",
    "dividend_income",
    "pension_income",
    "employment_expenses",
    "property_income",
    "gift_aid",
    "pension_contributions",
]


def impute_incomes(dataset: Dataset = FRS, year: int = 2022) -> MicroDataFrame:
    """Imputation of high incomes from the SPI.

    Args:
        dataset (type): The dataset to clone.
        year (int): The year to clone.

    Returns:
        Dict[str, ArrayLike]: The mapping from the original dataset to the cloned dataset.
    """
    from policyengine_uk import Microsimulation

    # Most recent SPI used - if it's before the FRS year then data will be uprated
    # automatically by OpenFisca-UK
    spi = Microsimulation(dataset=SPI)
    frs = Microsimulation(
        dataset=dataset,
        year=year,
    )

    regions = spi.calc("region").unique()

    spi_df = spi.df(PREDICTORS + IMPUTATIONS)
    frs_df = frs.df(PREDICTORS)

    frs_df.region = frs_df.region.map(
        {name: float(i) for i, name in enumerate(regions)}
    )
    spi_df.region = spi_df.region.map(
        {name: float(i) for i, name in enumerate(regions)}
    )

    return si.rf_impute(
        x_train=spi_df.drop(IMPUTATIONS, axis=1),
        y_train=spi_df[IMPUTATIONS],
        x_new=frs_df,
        verbose=True,
    )



import logging
from policyengine_uk import CountryTaxBenefitSystem
from policyengine_uk.data import EnhancedFRS
from policyengine_uk.entities import entities
import numpy as np
import warnings
from policyengine_uk.entities import *
import numpy as np
import warnings
from policyengine_uk.initial_setup import REPO, set_default
from policyengine_uk.reforms.presets.current_date import use_current_parameters
from policyengine_uk.reforms.presets.average_parameters import (
    average_parameters as apply_parameter_averaging,
)
from policyengine_uk.tools.baseline_variables import (
    generate_baseline_variables,
)
from policyengine_uk.tools.parameters import backdate_parameters
from policyengine_core.reforms import Reform
from policyengine_uk.data import DATASETS, SynthFRS
from policyengine_core.simulations import (
    Microsimulation as GeneralMicrosimulation,
    IndividualSim as GeneralIndividualSim,
)
import yaml
from pathlib import Path
import h5py
import pandas as pd
from policyengine_core.parameters.operations import (
    interpolate_parameters,
    uprate_parameters,
    propagate_parameter_metadata,
)
from policyengine_core.model_api import Reform
from policyengine_uk.tools.tax_benefit_uprating import add_tax_benefit_uprating
from functools import reduce


with open(Path(__file__).parent / "datasets.yml") as f:
    datasets = yaml.safe_load(f)
    DEFAULT_DATASET = list(
        filter(lambda ds: ds.name == datasets["default"], DATASETS)
    )[0]


warnings.filterwarnings("ignore")

np.random.seed(0)


def chain(*funcs):
    def f(parameters):
        for func in funcs:
            args = func(parameters)
        return args

    return f


class prepare_parameters(Reform):
    def apply(self):
        self.modify_parameters(
            chain(
                add_tax_benefit_uprating,
                propagate_parameter_metadata,
                interpolate_parameters,
                uprate_parameters,
            )
        )


class Microsimulation(GeneralMicrosimulation):
    tax_benefit_system = CountryTaxBenefitSystem
    entities = entities
    default_dataset = DEFAULT_DATASET
    post_reform = prepare_parameters, backdate_parameters()

    def __init__(
        self,
        reform: Reform = (),
        dataset: type = EnhancedFRS,
        year: int = 2022,
        adjust_weights: bool = False,
        average_parameters: bool = False,
        add_baseline_values: bool = True,
        post_reform: Reform = None,
    ):
        if len(dataset.years) == 0:
            logging.warning(
                f"You are trying to run a microsimulation using the dataset: {dataset.label}, but no years of that dataset could be found. Attempting to download it."
            )
            try:
                dataset.download(year)
            except Exception as e:
                logging.warning(
                    f"Encountered an error when attempting to download the {dataset.label} (this is likely because your account could not be authenticated, if it is not a public dataset). Attempting to download the Synthetic FRS."
                )
                try:
                    SynthFRS.download(year)
                    dataset = SynthFRS
                except Exception as e:
                    logging.warning(
                        f"Encountered an error when attempting to download the synthetic FRS dataset: {e}"
                    )
                    raise e
        if post_reform is not None:
            self.post_reform = post_reform
        if year is None:
            year = self.default_year or max(dataset.years)
        else:
            year = year
        year = int(year)

        # Check if dataset is available

        if year not in dataset.years:
            print(
                f"Dataset {dataset.name} does not contain year {year} (but it does contain {dataset.years}"
            )
            download = input(
                f"\nYear {year} not available in dataset {dataset.name}: \n\t* Download the dataset [y]\n\t* Use the synthetic FRS (and set default) [n]\n\nChoice: "
            )
            if download == "y":
                dataset.download(year)
            else:
                set_default(SynthFRS)
                dataset = SynthFRS
                if year not in dataset.years:
                    logging.info(
                        f"Year {year} synthetic FRS not stored, downloading..."
                    )
                    dataset.download(year)

        super().__init__(reform=reform, dataset=dataset, dataset_year=year)

        if (
            ("frs_enhanced" in dataset.name)
            and adjust_weights
            and year >= 2022
        ):
            weight_file = (
                Path(__file__).parent.parent / "calibration" / "frs_weights.h5"
            )
            if not weight_file.exists():
                raise FileNotFoundError("Weight adjustment file not found.")
            with h5py.File(weight_file, "r") as f:
                for year in f.keys():
                    self.simulation.set_input(
                        "household_weight", year, np.array(f[year])
                    )

                    self.simulation.set_input(
                        "person_weight",
                        year,
                        self.calc(
                            "household_weight", period=year, map_to="person"
                        ).values,
                    )

                    self.simulation.set_input(
                        "benunit_weight",
                        year,
                        self.calc(
                            "household_weight", period=year, map_to="benunit"
                        ).values,
                    )

        if average_parameters:
            self.simulation.tax_benefit_system.parameters = (
                apply_parameter_averaging(
                    self.simulation.tax_benefit_system.parameters
                )
            )


class IndividualSim(GeneralIndividualSim):
    tax_benefit_system = CountryTaxBenefitSystem
    post_reform = backdate_parameters()

    default_roles = dict(
        benunit="adult",
        household="adult",
    )
    required_entities = [
        "benunit",
        "household",
    ]



from policyengine_uk.model_api import *


class pension_credit_entitlement(Variable):
    label = "PC entitlement"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        gc = benunit("guarantee_credit", period)
        sc = benunit("savings_credit", period)
        eligible = benunit("is_pension_credit_eligible", period)
        return eligible * (gc + sc)



from policyengine_uk.model_api import *


class tax(Variable):
    value_type = float
    entity = Person
    label = "Taxes"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        TAXES = ["income_tax", "national_insurance"]
        return add(person, period, TAXES)


class household_tax(Variable):
    value_type = float
    entity = Household
    label = "taxes"
    documentation = "Total taxes owed by the household"
    definition_period = YEAR
    unit = GBP
    adds = [
        "change_in_expected_sdlt",
        "change_in_expected_ltt",
        "change_in_expected_lbtt",
        "corporate_sdlt_change_incidence",
        "business_rates_change_incidence",
        "council_tax",
        "domestic_rates",
        "change_in_fuel_duty",
        "tv_licence",
        "wealth_tax",
        "income_tax",
        "national_insurance",
        "LVT",
        "carbon_tax",
    ]


class benunit_tax(Variable):
    value_type = float
    entity = BenUnit
    label = "Benefit unit tax paid"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["tax"])


class tax_reported(Variable):
    value_type = float
    entity = Person
    label = "Reported tax paid"
    definition_period = YEAR
    unit = GBP


class tax_modelling(Variable):
    value_type = float
    entity = Person
    label = "Difference between reported and imputed tax liabilities"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        return person("tax", period) - person("tax_reported", period)



from policyengine_core.model_api import *
from policyengine_uk.entities import *
from policyengine_uk.tools.general import *
from policyengine_uk.tools.baseline_variables import (
    change_over_baseline,
    baseline_is_nonzero,
)
from microdf import MicroSeries, MicroDataFrame

GBP = "currency-GBP"



from policyengine_uk.variables.household.demographic.household import (
    TenureType,
)
from policyengine_uk.model_api import *


class benunit_id(Variable):
    value_type = int
    entity = BenUnit
    label = "ID for the family"
    definition_period = YEAR


class families(Variable):
    value_type = float
    entity = BenUnit
    label = "Variable holding families"
    definition_period = YEAR
    default_value = 1


class benunit_weight(Variable):
    value_type = float
    entity = BenUnit
    label = "Weight factor for the benefit unit"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.max(benunit.members("person_weight", period))


class is_married(Variable):
    value_type = bool
    entity = BenUnit
    label = "Married"
    documentation = "Whether the benefit unit adults are married to each other or in a civil partnership"
    definition_period = YEAR


class FamilyType(Enum):
    SINGLE = "Single, with no children"
    COUPLE_NO_CHILDREN = "Couple, with no children"
    LONE_PARENT = "Lone parent, with children"
    COUPLE_WITH_CHILDREN = "Couple, with children"


class RelationType(Enum):
    SINGLE = "Single"
    COUPLE = "Couple"


class family_type(Variable):
    value_type = Enum
    entity = BenUnit
    default_value = FamilyType.SINGLE
    possible_values = FamilyType
    label = "Family composition"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        two_adults = aggr(benunit, period, ["is_adult"]) == 2
        has_children = benunit.any(benunit.members("is_child", period))
        single = ~two_adults & ~has_children
        couple_no_children = two_adults & ~has_children
        lone_parent = ~two_adults & has_children
        couple_with_children = two_adults & has_children
        return select(
            [single, couple_no_children, lone_parent, couple_with_children],
            FamilyType,
        )


class relation_type(Variable):
    value_type = Enum
    entity = BenUnit
    default_value = RelationType.SINGLE
    possible_values = RelationType
    label = "Whether single or a couple"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return where(
            benunit.sum(benunit.members("is_adult", period)) == 1,
            RelationType.SINGLE,
            RelationType.COUPLE,
        )


class eldest_adult_age(Variable):
    value_type = float
    entity = BenUnit
    label = "Eldest adult age"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.max(
            where(
                benunit.members("is_adult", period),
                benunit.members("age", period),
                -np.inf,
            )
        )


class youngest_adult_age(Variable):
    value_type = float
    entity = BenUnit
    label = "Eldest adult age"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.min(
            where(
                benunit.members("is_adult", period),
                benunit.members("age", period),
                np.inf,
            )
        )


class eldest_child_age(Variable):
    value_type = float
    entity = BenUnit
    label = "Eldest adult age"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.max(
            where(
                benunit.members("is_child", period),
                benunit.members("age", period),
                -np.inf,
            )
        )


class youngest_child_age(Variable):
    value_type = float
    entity = BenUnit
    label = "Eldest adult age"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.min(
            where(
                benunit.members("is_child", period),
                benunit.members("age", period),
                np.inf,
            )
        )


class num_children(Variable):
    value_type = int
    entity = BenUnit
    label = "The number of children in the family"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.sum(benunit.members("is_child", period))


class num_adults(Variable):
    value_type = int
    entity = BenUnit
    label = "The number of adults in the family"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.sum(benunit.members("is_adult", period))


class benunit_tenure_type(Variable):
    value_type = Enum
    possible_values = TenureType
    default_value = TenureType.RENT_PRIVATELY
    entity = BenUnit
    label = "Tenure type of the family's household"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.value_from_first_person(
            benunit.members.household("tenure_type", period)
        )


class benunit_is_renting(Variable):
    value_type = bool
    entity = BenUnit
    label = "Whether this family is renting"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        tenure = benunit("benunit_tenure_type", period)
        tenures = tenure.possible_values
        RENT_TENURES = [
            tenures.RENT_PRIVATELY,
            tenures.RENT_FROM_COUNCIL,
            tenures.RENT_FROM_HA,
        ]
        return np.isin(tenure, RENT_TENURES)



from policyengine_uk.model_api import *
from policyengine_uk.variables.household.demographic.locations import BRMAName
from policyengine_uk.variables.household.demographic.geography import Region

label = "Geography"
index = -1


class BRMA(Variable):
    value_type = Enum
    possible_values = BRMAName
    default_value = BRMAName.MAIDSTONE
    entity = Household
    label = "Broad Rental Market Area"
    definition_period = YEAR


class region(Variable):
    value_type = Enum
    possible_values = Region
    default_value = Region.LONDON
    entity = Household
    label = "region"
    documentation = "Area of the UK"
    definition_period = YEAR



from policyengine_uk.model_api import *

label = "Transactions"


class property_purchased(Variable):
    label = "All property bought this year"
    documentation = "Whether all property wealth was bought this year"
    entity = Household
    definition_period = YEAR
    value_type = bool
    unit = GBP
    default_value = True


class cumulative_residential_rent(Variable):
    label = "Cumulative residential rent"
    documentation = "Total rent paid over the lifetime of the residential property a tenancy is held for."
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP


class cumulative_non_residential_rent(Variable):
    label = "Cumulative non-residential rent"
    documentation = "Total rent paid over the lifetime of the non-residential property a tenancy is held for."
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP


class rent(Variable):
    label = "Rent"
    documentation = (
        "The total amount of rent paid by the household in the year."
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    quantity_type = FLOW


class non_residential_rent(Variable):
    label = "Non-residential rent"
    documentation = "The total amount of rent paid by the household in the year for non-residential property."
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP



from policyengine_uk.model_api import *


class benunit_rent(Variable):
    value_type = float
    entity = BenUnit
    label = "Rent"
    documentation = "Gross rent that members of this family are liable for"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["personal_rent"])


class personal_rent(Variable):
    value_type = float
    entity = Person
    label = "Rent liable"
    documentation = "The gross rent this person is liable for"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        rent = person.household("rent", period)
        is_household_head = person("is_household_head", period)
        return rent * is_household_head


class family_rent(Variable):
    value_type = float
    entity = BenUnit
    label = "Gross rent for the family"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        personal_rent = benunit.members("personal_rent", period)
        return benunit.sum(personal_rent)


class weekly_childcare_expenses(Variable):
    value_type = float
    entity = Person
    label = "Average cost of childcare"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        return person("childcare_expenses", period) / WEEKS_IN_YEAR


class housing_costs(Variable):
    value_type = float
    entity = Household
    label = "Total housing costs"
    definition_period = YEAR
    unit = GBP

    def formula(household, period, parameters):
        return add(household, period, ["rent", "mortgage"])


class council_tax_less_benefit(Variable):
    label = "Council Tax (less CTB)"
    documentation = "Council Tax minus the Council Tax Benefit"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        person = household.members
        council_tax_benefit = household.sum(
            person.benunit("council_tax_benefit", period)
            * person("is_benunit_head", period)
        )
        return household("council_tax", period) - council_tax_benefit



from policyengine_uk.model_api import *

label = "Wealth"
description = "Wealth held by households."


class corporate_wealth(Variable):
    label = "corporate wealth"
    documentation = "Total owned wealth in corporations"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    quantity_type = STOCK


class owned_land(Variable):
    entity = Household
    label = "owned land value"
    documentation = "Total value of all land-only plots owned by the household"
    unit = GBP
    definition_period = YEAR
    value_type = float
    quantity_type = STOCK


class main_residence_value(Variable):
    label = "main residence value"
    documentation = "Total value of the main residence"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    quantity_type = STOCK


class other_residential_property_value(Variable):
    label = "other residence value"
    documentation = (
        "Total value of all residential property owned by the household"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    quantity_type = STOCK


class non_residential_property_value(Variable):
    label = "non-residential property value"
    documentation = (
        "Total value of all non-residential property owned by the household"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    quantity_type = STOCK



from policyengine_uk.model_api import *


class bi_maximum(Variable):
    label = "Basic income before phase-outs"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(person, period, parameters):
        bi = parameters(period).gov.contrib.ubi_center.basic_income
        weekly_flat_amount = bi.amount.flat
        is_senior_for_bi = person("is_SP_age", period)
        is_child_for_bi = person("age", period) < bi.amount.adult_age
        weekly_amount_by_age = select(
            [is_child_for_bi, is_senior_for_bi],
            [bi.amount.by_age.child, bi.amount.by_age.senior],
            default=bi.amount.by_age.working_age,
        )
        return (weekly_flat_amount + weekly_amount_by_age) * WEEKS_IN_YEAR



from policyengine_uk.model_api import *

label = "Energy"
description = "Energy consumption."


class domestic_energy_consumption(Variable):
    label = "Domestic energy consumption"
    documentation = "Combined gas and electric bills."
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    uprating = "gov.ofgem.price_cap.base"



from policyengine_core.data import Dataset
import h5py

from policyengine_uk.data.datasets.frs.enhanced.stages.imputation.enhanced_frs import (
    EnhancedFRS,
)


def remove_zero_weight_households(dataset: Dataset, year: int):
    """Removes zero-weight households (and benefit units and people) from a year of the given dataset.

    Args:
        dataset (Dataset): The dataset to edit.
        year (int): The year of the dataset to edit.
    """

    from policyengine_uk import Microsimulation

    sim = Microsimulation(dataset=dataset, dataset_year=year)

    # To be removed, households must have zero weight in all of these years
    YEARS = list(range(year, 2027))

    variables = dataset.keys(year)

    for variable in variables:
        if variable not in sim.tax_benefit_system.variables:
            continue
        entity = sim.tax_benefit_system.variables[variable].entity.key
        has_nonzero_weight = (
            sum(
                [
                    sim.calc(f"{entity}_weight", period=year).values
                    for year in YEARS
                ]
            )
            > 0
        )
        if dataset.data_format == Dataset.ARRAYS:
            dataset.save(
                year,
                variable,
                dataset.load(year, variable)[has_nonzero_weight],
            )
        elif dataset.data_format == Dataset.TIME_PERIOD_ARRAYS:
            for period in dataset.load(year, variable):
                key = f"{variable}/{period}"
                dataset.save(
                    year, key, dataset.load(year, key)[has_nonzero_weight]
                )


if __name__ == "__main__":
    remove_zero_weight_households(EnhancedFRS, 2022)



from policyengine_uk.model_api import *


class energy_bills_rebate(Variable):
    label = "Energy Bills Rebate"
    documentation = "Amount paid under the Energy Bills Rebate scheme."
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        return add(
            household,
            period,
            ["ebr_council_tax_rebate", "ebr_energy_bills_credit"],
        )



from policyengine_uk.model_api import *


class carbon_tax(Variable):
    entity = Household
    label = "Carbon tax"
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        carbon_tax = parameters(period).gov.contrib.ubi_center.carbon_tax
        rate = carbon_tax.rate
        emissions = household("carbon_consumption", period)
        # Household's share of total stocks and other corporate tax exposure.
        shareholding = household("shareholding", period)
        total_emissions = (
            emissions * household("household_weight", period)
        ).sum()
        consumer_incidence = carbon_tax.consumer_incidence * rate * emissions
        corporate_incidence = (
            (1 - carbon_tax.consumer_incidence)
            * rate
            * total_emissions
            * shareholding
        )
        return consumer_incidence + corporate_incidence



from policyengine_uk.model_api import *

# The below variables follow the COICOP MECE categories.

label = "General"
description = "General consumption categories"


class food_and_non_alcoholic_beverages_consumption(Variable):
    entity = Household
    label = "food and alcoholic beverage consumption"
    documentation = "Total yearly expenditure on food and alcoholic beverages"
    unit = GBP
    definition_period = YEAR
    value_type = float
    quantity_type = FLOW
    uprating = "calibration.uprating.monthly_cpi_by_category.food_and_non_alcoholic_beverages"


class alcohol_and_tobacco_consumption(Variable):
    entity = Household
    label = "alcohol and tobacco consumption"
    documentation = "Total yearly expenditure on alcohol and tobacco"
    unit = GBP
    definition_period = YEAR
    value_type = float
    quantity_type = FLOW
    uprating = (
        "calibration.uprating.monthly_cpi_by_category.alcohol_and_tobacco"
    )


class clothing_and_footwear_consumption(Variable):
    entity = Household
    label = "clothing and footwear consumption"
    documentation = "Total yearly expenditure on clothing and footwear"
    unit = GBP
    definition_period = YEAR
    value_type = float
    quantity_type = FLOW
    uprating = (
        "calibration.uprating.monthly_cpi_by_category.clothing_and_footwear"
    )


class housing_water_and_electricity_consumption(Variable):
    entity = Household
    label = "housing, water and electricity consumption"
    documentation = (
        "Total yearly expenditure on housing, water and electricity"
    )
    unit = GBP
    definition_period = YEAR
    value_type = float
    quantity_type = FLOW
    uprating = "calibration.uprating.monthly_cpi_by_category.housing_water_and_electricity"


class household_furnishings_consumption(Variable):
    entity = Household
    label = "household furnishings consumption"
    documentation = "Total yearly expenditure on household furnishings"
    unit = GBP
    definition_period = YEAR
    value_type = float
    quantity_type = FLOW


class health_consumption(Variable):
    entity = Household
    label = "health consumption"
    documentation = "Total yearly expenditure on health"
    unit = GBP
    definition_period = YEAR
    value_type = float
    quantity_type = FLOW
    uprating = "calibration.uprating.monthly_cpi_by_category.health"


class transport_consumption(Variable):
    entity = Household
    label = "transport consumption"
    documentation = "Total yearly expenditure on transport"
    unit = GBP
    definition_period = YEAR
    value_type = float
    quantity_type = FLOW
    uprating = "calibration.uprating.monthly_cpi_by_category.transport"


class communication_consumption(Variable):
    entity = Household
    label = "communication consumption"
    documentation = "Total yearly expenditure on communication"
    unit = GBP
    definition_period = YEAR
    value_type = float
    quantity_type = FLOW
    uprating = "calibration.uprating.monthly_cpi_by_category.communication"


class recreation_consumption(Variable):
    entity = Household
    label = "recreation consumption"
    documentation = "Total yearly expenditure on recreation"
    unit = GBP
    definition_period = YEAR
    value_type = float
    quantity_type = FLOW
    uprating = "calibration.uprating.monthly_cpi_by_category.recreation"


class education_consumption(Variable):
    entity = Household
    label = "education consumption"
    documentation = "Total yearly expenditure on education"
    unit = GBP
    definition_period = YEAR
    value_type = float
    quantity_type = FLOW
    uprating = "calibration.uprating.monthly_cpi_by_category.education"


class restaurants_and_hotels_consumption(Variable):
    entity = Household
    label = "restaurants and hotels consumption"
    documentation = "Total yearly expenditure on restaurants and hotels"
    unit = GBP
    definition_period = YEAR
    value_type = float
    quantity_type = FLOW
    uprating = (
        "calibration.uprating.monthly_cpi_by_category.restaurants_and_hotels"
    )


class miscellaneous_consumption(Variable):
    entity = Household
    label = "miscellaneous consumption"
    documentation = "Total yearly expenditure on miscellaneous goods"
    unit = GBP
    definition_period = YEAR
    value_type = float
    quantity_type = FLOW
    uprating = "calibration.uprating.monthly_cpi_by_category.miscellaneous"


class petrol_spending(Variable):
    label = "petrol consumption"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    quantity_type = FLOW
    uprating = "calibration.uprating.monthly_cpi_by_category.transport"


class diesel_spending(Variable):
    label = "diesel consumption"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    quantity_type = FLOW
    uprating = "calibration.uprating.monthly_cpi_by_category.transport"


class childcare_expenses(Variable):
    value_type = float
    entity = Person
    label = "childcare consumption"
    documentation = "Total amount spent on childcare"
    definition_period = YEAR
    unit = GBP


class private_pension_contributions(Variable):
    value_type = float
    entity = Person
    label = "private pension contributions"
    documentation = "Total amount spent on private pension contributions"
    definition_period = YEAR
    unit = GBP


class occupational_pension_contributions(Variable):
    value_type = float
    entity = Person
    label = "occupational pension contributions"
    documentation = "Total amount spent on occupational pension contributions"
    definition_period = YEAR
    unit = GBP



from policyengine_uk.model_api import *


class person_benunit_id(Variable):
    value_type = float
    label = "Person's benefit unit ID"
    entity = Person
    definition_period = YEAR


class person_household_id(Variable):
    value_type = float
    label = "Person's household ID"
    entity = Person
    definition_period = YEAR


class role(Variable):
    value_type = str
    label = "Role (adult/child)"
    entity = Person
    definition_period = YEAR


class person_benunit_role(Variable):
    value_type = str
    label = "Role (adult/child)"
    entity = Person
    definition_period = YEAR


class person_household_role(Variable):
    value_type = str
    label = "Role (adult/child)"
    entity = Person
    definition_period = YEAR



from policyengine_uk.model_api import *


class BRMAName(Enum):
    ASHFORD = "Ashford"
    AYLESBURY = "Aylesbury"
    BARNSLEY = "Barnsley"
    BARROW_IN_FURNESS = "Barrow-in-Furness"
    BASINGSTOKE = "Basingstoke"
    BATH = "Bath"
    BEDFORD = "Bedford"
    BIRMINGHAM = "Birmingham"
    BLACK_COUNTRY = "Black Country"
    BLACKWATER_VALLEY = "Blackwater Valley"
    BOLTON_AND_BURY = "Bolton and Bury"
    BOURNEMOUTH = "Bournemouth"
    BRADFORD_SOUTH_DALES = "Bradford & South Dales"
    BRIGHTON_AND_HOVE = "Brighton and Hove"
    BRISTOL = "Bristol"
    BURY_ST_EDMUNDS = "Bury St Edmunds"
    CAMBRIDGE = "Cambridge"
    CANTERBURY = "Canterbury"
    CENTRAL_GREATER_MANCHESTER = "Central Greater Manchester"
    CENTRAL_LANCS = "Central Lancs"
    CENTRAL_LONDON = "Central London"
    CENTRAL_NORFOLK_NORWICH = "Central Norfolk & Norwich"
    CHELMSFORD = "Chelmsford"
    CHELTENHAM = "Cheltenham"
    CHERWELL_VALLEY = "Cherwell Valley"
    CHESTERFIELD = "Chesterfield"
    CHICHESTER = "Chichester"
    CHILTERNS = "Chilterns"
    COLCHESTER = "Colchester"
    COVENTRY = "Coventry"
    CRAWLEY_REIGATE = "Crawley & Reigate"
    DARLINGTON = "Darlington"
    DERBY = "Derby"
    DONCASTER = "Doncaster"
    DOVER_SHEPWAY = "Dover-Shepway"
    DURHAM = "Durham"
    EAST_CHESHIRE = "East Cheshire"
    EAST_LANCS = "East Lancs"
    EAST_THAMES_VALLEY = "East Thames Valley"
    EASTBOURNE = "Eastbourne"
    EASTERN_STAFFORDSHIRE = "Eastern Staffordshire"
    EXETER = "Exeter"
    FYLDE_COAST = "Fylde Coast"
    GLOUCESTER = "Gloucester"
    GRANTHAM_NEWARK = "Grantham & Newark"
    GREATER_LIVERPOOL = "Greater Liverpool"
    GRIMSBY = "Grimsby"
    GUILDFORD = "Guildford"
    HALIFAX = "Halifax"
    HARLOW_STORTFORD = "Harlow & Stortford"
    HARROGATE = "Harrogate"
    HEREFORDSHIRE = "Herefordshire"
    HIGH_WEALD = "High Weald"
    HULL_EAST_RIDING = "Hull & East Riding"
    HUNTINGDON = "Huntingdon"
    INNER_EAST_LONDON = "Inner East London"
    INNER_NORTH_LONDON = "Inner North London"
    INNER_SOUTH_EAST_LONDON = "Inner South East London"
    INNER_SOUTH_WEST_LONDON = "Inner South West London"
    INNER_WEST_LONDON = "Inner West London"
    IPSWICH = "Ipswich"
    ISLE_OF_WIGHT = "Isle of Wight"
    KENDAL = "Kendal"
    KERNOW_WEST = "Kernow West"
    KINGS_LYNN = "Kings Lynn"
    KIRKLEES = "Kirklees"
    LANCASTER = "Lancaster"
    LEEDS = "Leeds"
    LEICESTER = "Leicester"
    LINCOLN = "Lincoln"
    LINCOLNSHIRE_FENS = "Lincolnshire Fens"
    LOWESTOFT_GREAT_YARMOUTH = "Lowestoft & Great Yarmouth"
    LUTON = "Luton"
    MAIDSTONE = "Maidstone"
    MEDWAY_SWALE = "Medway & Swale"
    MENDIP = "Mendip"
    MID_EAST_DEVON = "Mid & East Devon"
    MID_WEST_DORSET = "Mid & West Dorset"
    MID_STAFFS = "Mid Staffs"
    MILTON_KEYNES = "Milton Keynes"
    NEWBURY = "Newbury"
    NORTH_CHESHIRE = "North Cheshire"
    NORTH_CORNWALL_DEVON_BORDERS = "North Cornwall & Devon Borders"
    NORTH_CUMBRIA = "North Cumbria"
    NORTH_DEVON = "North Devon"
    NORTH_NOTTINGHAM = "North Nottingham"
    NORTH_WEST_KENT = "North West Kent"
    NORTH_WEST_LONDON = "North West London"
    NORTHAMPTON = "Northampton"
    NORTHANTS_CENTRAL = "Northants Central"
    NORTHUMBERLAND = "Northumberland"
    NOTTINGHAM = "Nottingham"
    OLDHAM_ROCHDALE = "Oldham & Rochdale"
    OUTER_EAST_LONDON = "Outer East London"
    OUTER_NORTH_EAST_LONDON = "Outer North East London"
    OUTER_NORTH_LONDON = "Outer North London"
    OUTER_SOUTH_EAST_LONDON = "Outer South East London"
    OUTER_SOUTH_LONDON = "Outer South London"
    OUTER_SOUTH_WEST_LONDON = "Outer South West London"
    OUTER_WEST_LONDON = "Outer West London"
    OXFORD = "Oxford"
    PEAKS_DALES = "Peaks & Dales"
    PETERBOROUGH = "Peterborough"
    PLYMOUTH = "Plymouth"
    PORTSMOUTH = "Portsmouth"
    READING = "Reading"
    RICHMOND_HAMBLETON = "Richmond & Hambleton"
    ROTHERHAM = "Rotherham"
    RUGBY_EAST = "Rugby & East"
    SALISBURY = "Salisbury"
    SCARBOROUGH = "Scarborough"
    SCUNTHORPE = "Scunthorpe"
    SHEFFIELD = "Sheffield"
    SHROPSHIRE = "Shropshire"
    SOLIHULL = "Solihull"
    SOUTH_CHESHIRE = "South Cheshire"
    SOUTH_DEVON = "South Devon"
    SOUTH_EAST_HERTS = "South East Herts"
    SOUTH_WEST_ESSEX = "South West Essex"
    SOUTH_WEST_HERTS = "South West Herts"
    SOUTHAMPTON = "Southampton"
    SOUTHEND = "Southend"
    SOUTHERN_GREATER_MANCHESTER = "Southern Greater Manchester"
    SOUTHPORT = "Southport"
    ST_HELENS = "St Helens"
    STAFFORDSHIRE_NORTH = "Staffordshire North"
    STEVENAGE_NORTH_HERTS = "Stevenage & North Herts"
    SUNDERLAND = "Sunderland"
    SUSSEX_EAST = "Sussex East"
    SWINDON = "Swindon"
    TAMESIDE_GLOSSOP = "Tameside & Glossop"
    TAUNTON_WEST_SOMERSET = "Taunton & West Somerset"
    TEESSIDE = "Teesside"
    THANET = "Thanet"
    TYNESIDE = "Tyneside"
    WAKEFIELD = "Wakefield"
    WALTON = "Walton"
    WARWICKSHIRE_SOUTH = "Warwickshire South"
    WEST_CHESHIRE = "West Cheshire"
    WEST_CUMBRIA = "West Cumbria"
    WEST_PENNINE = "West Pennine"
    WEST_WILTSHIRE = "West Wiltshire"
    WESTON_S_MARE = "Weston-S-Mare"
    WIGAN = "Wigan"
    WINCHESTER = "Winchester"
    WIRRAL = "Wirral"
    WOLDS_AND_COAST = "Wolds and Coast"
    WORCESTER_NORTH = "Worcester North"
    WORCESTER_SOUTH = "Worcester South"
    WORTHING = "Worthing"
    YEOVIL = "Yeovil"
    YORK = "York"
    # Northern Ireland
    BELFAST = "Belfast"
    LOUGH_NEAGH_LOWER = "Lough Neagh Lower"
    LOUGH_NEAGH_UPPER = "Lough Neagh Upper"
    NORTH = "North (NI)"
    NORTH_WEST = "North West (NI)"
    SOUTH = "South (NI)"
    SOUTH_EAST = "South East (NI)"
    SOUTH_WEST = "South West (NI)"


class LocalAuthority(Enum):
    ABERDEEN_CITY = "Aberdeen City"
    ABERDEENSHIRE = "Aberdeenshire"
    ADUR = "Adur"
    ALLERDALE = "Allerdale"
    AMBER_VALLEY = "Amber Valley"
    ANGUS = "Angus"
    ANTRIM_AND_NEWTOWNABBEY = "Antrim and Newtownabbey"
    ARDS_AND_NORTH_DOWN = "Ards and North Down"
    ARGYLL_AND_BUTE = "Argyll and Bute"
    ARMAGH_CITY_BANBRIDGE_AND_CRAIGAVON = (
        "Armagh City, Banbridge and Craigavon"
    )
    ARUN = "Arun"
    ASHFIELD = "Ashfield"
    ASHFORD = "Ashford"
    BABERGH = "Babergh"
    BARKING_AND_DAGENHAM = "Barking and Dagenham"
    BARNET = "Barnet"
    BARNSLEY = "Barnsley"
    BARROW_IN_FURNESS = "Barrow-in-Furness"
    BASILDON = "Basildon"
    BASINGSTOKE_AND_DEANE = "Basingstoke and Deane"
    BASSETLAW = "Bassetlaw"
    BATH_AND_NORTH_EAST_SOMERSET = "Bath and North East Somerset"
    BEDFORD = "Bedford"
    BELFAST = "Belfast"
    BEXLEY = "Bexley"
    BIRMINGHAM = "Birmingham"
    BLABY = "Blaby"
    BLACKBURN_WITH_DARWEN = "Blackburn with Darwen"
    BLACKPOOL = "Blackpool"
    BLAENAU_GWENT = "Blaenau Gwent"
    BOLSOVER = "Bolsover"
    BOLTON = "Bolton"
    BOSTON = "Boston"
    BOURNEMOUTH_CHRISTCHURCH_AND_POOLE = "Bournemouth, Christchurch and Poole"
    BRACKNELL_FOREST = "Bracknell Forest"
    BRADFORD = "Bradford"
    BRAINTREE = "Braintree"
    BRECKLAND = "Breckland"
    BRENT = "Brent"
    BRENTWOOD = "Brentwood"
    BRIDGEND = "Bridgend"
    BRIGHTON_AND_HOVE = "Brighton and Hove"
    BRISTOL_CITY_OF = "Bristol, City of"
    BROADLAND = "Broadland"
    BROMLEY = "Bromley"
    BROMSGROVE = "Bromsgrove"
    BROXBOURNE = "Broxbourne"
    BROXTOWE = "Broxtowe"
    BUCKINGHAMSHIRE = "Buckinghamshire"
    BURNLEY = "Burnley"
    BURY = "Bury"
    CAERPHILLY = "Caerphilly"
    CALDERDALE = "Calderdale"
    CAMBRIDGE = "Cambridge"
    CAMDEN = "Camden"
    CANNOCK_CHASE = "Cannock Chase"
    CANTERBURY = "Canterbury"
    CARDIFF = "Cardiff"
    CARLISLE = "Carlisle"
    CARMARTHENSHIRE = "Carmarthenshire"
    CASTLE_POINT = "Castle Point"
    CAUSEWAY_COAST_AND_GLENS = "Causeway Coast and Glens"
    CENTRAL_BEDFORDSHIRE = "Central Bedfordshire"
    CEREDIGION = "Ceredigion"
    CHARNWOOD = "Charnwood"
    CHELMSFORD = "Chelmsford"
    CHELTENHAM = "Cheltenham"
    CHERWELL = "Cherwell"
    CHESHIRE_EAST = "Cheshire East"
    CHESHIRE_WEST_AND_CHESTER = "Cheshire West and Chester"
    CHESTERFIELD = "Chesterfield"
    CHICHESTER = "Chichester"
    CHORLEY = "Chorley"
    CITY_OF_EDINBURGH = "City of Edinburgh"
    CITY_OF_LONDON = "City of London"
    CLACKMANNANSHIRE = "Clackmannanshire"
    COLCHESTER = "Colchester"
    CONWY = "Conwy"
    COPELAND = "Copeland"
    CORBY = "Corby"
    CORNWALL = "Cornwall"
    COTSWOLD = "Cotswold"
    COUNTY_DURHAM = "County Durham"
    COVENTRY = "Coventry"
    CRAVEN = "Craven"
    CRAWLEY = "Crawley"
    CROYDON = "Croydon"
    DACORUM = "Dacorum"
    DARLINGTON = "Darlington"
    DARTFORD = "Dartford"
    DAVENTRY = "Daventry"
    DENBIGHSHIRE = "Denbighshire"
    DERBY = "Derby"
    DERBYSHIRE_DALES = "Derbyshire Dales"
    DERRY_CITY_AND_STRABANE = "Derry City and Strabane"
    DONCASTER = "Doncaster"
    DORSET = "Dorset"
    DOVER = "Dover"
    DUDLEY = "Dudley"
    DUMFRIES_AND_GALLOWAY = "Dumfries and Galloway"
    DUNDEE_CITY = "Dundee City"
    EALING = "Ealing"
    EAST_AYRSHIRE = "East Ayrshire"
    EAST_CAMBRIDGESHIRE = "East Cambridgeshire"
    EAST_DEVON = "East Devon"
    EAST_DUNBARTONSHIRE = "East Dunbartonshire"
    EAST_HAMPSHIRE = "East Hampshire"
    EAST_HERTFORDSHIRE = "East Hertfordshire"
    EAST_LINDSEY = "East Lindsey"
    EAST_LOTHIAN = "East Lothian"
    EAST_NORTHAMPTONSHIRE = "East Northamptonshire"
    EAST_RENFREWSHIRE = "East Renfrewshire"
    EAST_RIDING_OF_YORKSHIRE = "East Riding of Yorkshire"
    EAST_STAFFORDSHIRE = "East Staffordshire"
    EAST_SUFFOLK = "East Suffolk"
    EASTBOURNE = "Eastbourne"
    EASTLEIGH = "Eastleigh"
    EDEN = "Eden"
    ELMBRIDGE = "Elmbridge"
    ENFIELD = "Enfield"
    EPPING_FOREST = "Epping Forest"
    EPSOM_AND_EWELL = "Epsom and Ewell"
    EREWASH = "Erewash"
    EXETER = "Exeter"
    FALKIRK = "Falkirk"
    FAREHAM = "Fareham"
    FENLAND = "Fenland"
    FERMANAGH_AND_OMAGH = "Fermanagh and Omagh"
    FIFE = "Fife"
    FLINTSHIRE = "Flintshire"
    FOLKESTONE_AND_HYTHE = "Folkestone and Hythe"
    FOREST_OF_DEAN = "Forest of Dean"
    FYLDE = "Fylde"
    GATESHEAD = "Gateshead"
    GEDLING = "Gedling"
    GLASGOW_CITY = "Glasgow City"
    GLOUCESTER = "Gloucester"
    GOSPORT = "Gosport"
    GRAVESHAM = "Gravesham"
    GREAT_YARMOUTH = "Great Yarmouth"
    GREENWICH = "Greenwich"
    GUILDFORD = "Guildford"
    GWYNEDD = "Gwynedd"
    HACKNEY = "Hackney"
    HALTON = "Halton"
    HAMBLETON = "Hambleton"
    HAMMERSMITH_AND_FULHAM = "Hammersmith and Fulham"
    HARBOROUGH = "Harborough"
    HARINGEY = "Haringey"
    HARLOW = "Harlow"
    HARROGATE = "Harrogate"
    HARROW = "Harrow"
    HART = "Hart"
    HARTLEPOOL = "Hartlepool"
    HASTINGS = "Hastings"
    HAVANT = "Havant"
    HAVERING = "Havering"
    HEREFORDSHIRE_COUNTY_OF = "Herefordshire, County of"
    HERTSMERE = "Hertsmere"
    HIGH_PEAK = "High Peak"
    HIGHLAND = "Highland"
    HILLINGDON = "Hillingdon"
    HINCKLEY_AND_BOSWORTH = "Hinckley and Bosworth"
    HORSHAM = "Horsham"
    HOUNSLOW = "Hounslow"
    HUNTINGDONSHIRE = "Huntingdonshire"
    HYNDBURN = "Hyndburn"
    INVERCLYDE = "Inverclyde"
    IPSWICH = "Ipswich"
    ISLE_OF_ANGLESEY = "Isle of Anglesey"
    ISLE_OF_WIGHT = "Isle of Wight"
    ISLES_OF_SCILLY = "Isles of Scilly"
    ISLINGTON = "Islington"
    KENSINGTON_AND_CHELSEA = "Kensington and Chelsea"
    KETTERING = "Kettering"
    KINGS_LYNN_AND_WEST_NORFOLK = "King's Lynn and West Norfolk"
    KINGSTON_UPON_HULL_CITY_OF = "Kingston upon Hull, City of"
    KINGSTON_UPON_THAMES = "Kingston upon Thames"
    KIRKLEES = "Kirklees"
    KNOWSLEY = "Knowsley"
    LAMBETH = "Lambeth"
    LANCASTER = "Lancaster"
    LEEDS = "Leeds"
    LEICESTER = "Leicester"
    LEWES = "Lewes"
    LEWISHAM = "Lewisham"
    LICHFIELD = "Lichfield"
    LINCOLN = "Lincoln"
    LISBURN_AND_CASTLEREAGH = "Lisburn and Castlereagh"
    LIVERPOOL = "Liverpool"
    LUTON = "Luton"
    MAIDSTONE = "Maidstone"
    MALDON = "Maldon"
    MALVERN_HILLS = "Malvern Hills"
    MANCHESTER = "Manchester"
    MANSFIELD = "Mansfield"
    MEDWAY = "Medway"
    MELTON = "Melton"
    MENDIP = "Mendip"
    MERTHYR_TYDFIL = "Merthyr Tydfil"
    MERTON = "Merton"
    MID_DEVON = "Mid Devon"
    MID_SUFFOLK = "Mid Suffolk"
    MID_SUSSEX = "Mid Sussex"
    MID_ULSTER = "Mid Ulster"
    MID_AND_EAST_ANTRIM = "Mid and East Antrim"
    MIDDLESBROUGH = "Middlesbrough"
    MIDLOTHIAN = "Midlothian"
    MILTON_KEYNES = "Milton Keynes"
    MOLE_VALLEY = "Mole Valley"
    MONMOUTHSHIRE = "Monmouthshire"
    MORAY = "Moray"
    NA_H_EILEANAN_SIAR = "Na h-Eileanan Siar"
    NEATH_PORT_TALBOT = "Neath Port Talbot"
    NEW_FOREST = "New Forest"
    NEWARK_AND_SHERWOOD = "Newark and Sherwood"
    NEWCASTLE_UPON_TYNE = "Newcastle upon Tyne"
    NEWCASTLE_UNDER_LYME = "Newcastle-under-Lyme"
    NEWHAM = "Newham"
    NEWPORT = "Newport"
    NEWRY_MOURNE_AND_DOWN = "Newry, Mourne and Down"
    NORTH_AYRSHIRE = "North Ayrshire"
    NORTH_DEVON = "North Devon"
    NORTH_EAST_DERBYSHIRE = "North East Derbyshire"
    NORTH_EAST_LINCOLNSHIRE = "North East Lincolnshire"
    NORTH_HERTFORDSHIRE = "North Hertfordshire"
    NORTH_KESTEVEN = "North Kesteven"
    NORTH_LANARKSHIRE = "North Lanarkshire"
    NORTH_LINCOLNSHIRE = "North Lincolnshire"
    NORTH_NORFOLK = "North Norfolk"
    NORTH_SOMERSET = "North Somerset"
    NORTH_TYNESIDE = "North Tyneside"
    NORTH_WARWICKSHIRE = "North Warwickshire"
    NORTH_WEST_LEICESTERSHIRE = "North West Leicestershire"
    NORTHAMPTON = "Northampton"
    NORTHUMBERLAND = "Northumberland"
    NORWICH = "Norwich"
    NOTTINGHAM = "Nottingham"
    NUNEATON_AND_BEDWORTH = "Nuneaton and Bedworth"
    OADBY_AND_WIGSTON = "Oadby and Wigston"
    OLDHAM = "Oldham"
    ORKNEY_ISLANDS = "Orkney Islands"
    OXFORD = "Oxford"
    PEMBROKESHIRE = "Pembrokeshire"
    PENDLE = "Pendle"
    PERTH_AND_KINROSS = "Perth and Kinross"
    PETERBOROUGH = "Peterborough"
    PLYMOUTH = "Plymouth"
    PORTSMOUTH = "Portsmouth"
    POWYS = "Powys"
    PRESTON = "Preston"
    READING = "Reading"
    REDBRIDGE = "Redbridge"
    REDCAR_AND_CLEVELAND = "Redcar and Cleveland"
    REDDITCH = "Redditch"
    REIGATE_AND_BANSTEAD = "Reigate and Banstead"
    RENFREWSHIRE = "Renfrewshire"
    RHONDDA_CYNON_TAF = "Rhondda Cynon Taf"
    RIBBLE_VALLEY = "Ribble Valley"
    RICHMOND_UPON_THAMES = "Richmond upon Thames"
    RICHMONDSHIRE = "Richmondshire"
    ROCHDALE = "Rochdale"
    ROCHFORD = "Rochford"
    ROSSENDALE = "Rossendale"
    ROTHER = "Rother"
    ROTHERHAM = "Rotherham"
    RUGBY = "Rugby"
    RUNNYMEDE = "Runnymede"
    RUSHCLIFFE = "Rushcliffe"
    RUSHMOOR = "Rushmoor"
    RUTLAND = "Rutland"
    RYEDALE = "Ryedale"
    SALFORD = "Salford"
    SANDWELL = "Sandwell"
    SCARBOROUGH = "Scarborough"
    SCOTTISH_BORDERS = "Scottish Borders"
    SEDGEMOOR = "Sedgemoor"
    SEFTON = "Sefton"
    SELBY = "Selby"
    SEVENOAKS = "Sevenoaks"
    SHEFFIELD = "Sheffield"
    SHETLAND_ISLANDS = "Shetland Islands"
    SHROPSHIRE = "Shropshire"
    SLOUGH = "Slough"
    SOLIHULL = "Solihull"
    SOMERSET_WEST_AND_TAUNTON = "Somerset West and Taunton"
    SOUTH_AYRSHIRE = "South Ayrshire"
    SOUTH_CAMBRIDGESHIRE = "South Cambridgeshire"
    SOUTH_DERBYSHIRE = "South Derbyshire"
    SOUTH_GLOUCESTERSHIRE = "South Gloucestershire"
    SOUTH_HAMS = "South Hams"
    SOUTH_HOLLAND = "South Holland"
    SOUTH_KESTEVEN = "South Kesteven"
    SOUTH_LAKELAND = "South Lakeland"
    SOUTH_LANARKSHIRE = "South Lanarkshire"
    SOUTH_NORFOLK = "South Norfolk"
    SOUTH_NORTHAMPTONSHIRE = "South Northamptonshire"
    SOUTH_OXFORDSHIRE = "South Oxfordshire"
    SOUTH_RIBBLE = "South Ribble"
    SOUTH_SOMERSET = "South Somerset"
    SOUTH_STAFFORDSHIRE = "South Staffordshire"
    SOUTH_TYNESIDE = "South Tyneside"
    SOUTHAMPTON = "Southampton"
    SOUTHEND_ON_SEA = "Southend-on-Sea"
    SOUTHWARK = "Southwark"
    SPELTHORNE = "Spelthorne"
    ST_ALBANS = "St Albans"
    ST_HELENS = "St. Helens"
    STAFFORD = "Stafford"
    STAFFORDSHIRE_MOORLANDS = "Staffordshire Moorlands"
    STEVENAGE = "Stevenage"
    STIRLING = "Stirling"
    STOCKPORT = "Stockport"
    STOCKTON_ON_TEES = "Stockton-on-Tees"
    STOKE_ON_TRENT = "Stoke-on-Trent"
    STRATFORD_ON_AVON = "Stratford-on-Avon"
    STROUD = "Stroud"
    SUNDERLAND = "Sunderland"
    SURREY_HEATH = "Surrey Heath"
    SUTTON = "Sutton"
    SWALE = "Swale"
    SWANSEA = "Swansea"
    SWINDON = "Swindon"
    TAMESIDE = "Tameside"
    TAMWORTH = "Tamworth"
    TANDRIDGE = "Tandridge"
    TEIGNBRIDGE = "Teignbridge"
    TELFORD_AND_WREKIN = "Telford and Wrekin"
    TENDRING = "Tendring"
    TEST_VALLEY = "Test Valley"
    TEWKESBURY = "Tewkesbury"
    THANET = "Thanet"
    THREE_RIVERS = "Three Rivers"
    THURROCK = "Thurrock"
    TONBRIDGE_AND_MALLING = "Tonbridge and Malling"
    TORBAY = "Torbay"
    TORFAEN = "Torfaen"
    TORRIDGE = "Torridge"
    TOWER_HAMLETS = "Tower Hamlets"
    TRAFFORD = "Trafford"
    TUNBRIDGE_WELLS = "Tunbridge Wells"
    UTTLESFORD = "Uttlesford"
    VALE_OF_GLAMORGAN = "Vale of Glamorgan"
    VALE_OF_WHITE_HORSE = "Vale of White Horse"
    WAKEFIELD = "Wakefield"
    WALSALL = "Walsall"
    WALTHAM_FOREST = "Waltham Forest"
    WANDSWORTH = "Wandsworth"
    WARRINGTON = "Warrington"
    WARWICK = "Warwick"
    WATFORD = "Watford"
    WAVERLEY = "Waverley"
    WEALDEN = "Wealden"
    WELLINGBOROUGH = "Wellingborough"
    WELWYN_HATFIELD = "Welwyn Hatfield"
    WEST_BERKSHIRE = "West Berkshire"
    WEST_DEVON = "West Devon"
    WEST_DUNBARTONSHIRE = "West Dunbartonshire"
    WEST_LANCASHIRE = "West Lancashire"
    WEST_LINDSEY = "West Lindsey"
    WEST_LOTHIAN = "West Lothian"
    WEST_OXFORDSHIRE = "West Oxfordshire"
    WEST_SUFFOLK = "West Suffolk"
    WESTMINSTER = "Westminster"
    WIGAN = "Wigan"
    WILTSHIRE = "Wiltshire"
    WINCHESTER = "Winchester"
    WINDSOR_AND_MAIDENHEAD = "Windsor and Maidenhead"
    WIRRAL = "Wirral"
    WOKING = "Woking"
    WOKINGHAM = "Wokingham"
    WOLVERHAMPTON = "Wolverhampton"
    WORCESTER = "Worcester"
    WORTHING = "Worthing"
    WREXHAM = "Wrexham"
    WYCHAVON = "Wychavon"
    WYRE = "Wyre"
    WYRE_FOREST = "Wyre Forest"
    YORK = "York"


class local_authority(Variable):
    value_type = Enum
    possible_values = LocalAuthority
    default_value = LocalAuthority.MAIDSTONE
    entity = Household
    label = "Local Authority"
    documentation = "The local authority for this household (this is often the same as your Broad Rental Market Area, but may differ)."
    definition_period = YEAR



from policyengine_uk.model_api import *

label = "Demographic"
description = "Demographic variables."


class age(Variable):
    value_type = float
    entity = Person
    label = "age"
    unit = "year"
    documentation = "Age in years"
    definition_period = YEAR
    quantity_type = STOCK
    default_value = 18



from policyengine_uk.model_api import *


class count_children_and_qyp(Variable):
    label = "Children and qualifying young people"
    documentation = "The number of children and qualifying young people (young adults in education) in the family"
    entity = BenUnit
    definition_period = YEAR
    value_type = int
    unit = GBP

    def formula(benunit, period, parameters):
        return add(benunit, period, ["is_child_or_QYP"])



from policyengine_uk.model_api import *


class additional_minimum_guarantee(Variable):
    label = "Additional Minimum Guarantee"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/uksi/2002/1792/regulation/6"

    formula = sum_of_variables(
        "gov.dwp.pension_credit.guarantee_credit.additions"
    )



from policyengine_uk.model_api import *


class is_pension_credit_eligible(Variable):
    label = "Eligible for Pension Credit"
    entity = BenUnit
    definition_period = YEAR
    value_type = bool
    unit = GBP
    reference = "https://www.legislation.gov.uk/ukpga/2002/16/section/1"

    def formula(benunit, period, parameters):
        has_sp_age_member = benunit.any(benunit.members("is_SP_age", period))
        is_gc_eligible = benunit("is_guarantee_credit_eligible", period)
        is_sc_eligible = benunit("is_savings_credit_eligible", period)
        return has_sp_age_member & (is_gc_eligible | is_sc_eligible)



from policyengine_uk.model_api import *


class minimum_guarantee(Variable):
    label = "Minimum Guarantee"
    documentation = "The Minimum Guarantee informs the Pension Credit Guarantee Credit amount. Pensioners usually have their annual income increased to this amount by Pension Credit."
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/uksi/2002/1792/regulation/6"

    formula = sum_of_variables(
        [
            "standard_minimum_guarantee",
            "additional_minimum_guarantee",
        ]
    )



from policyengine_uk.model_api import *


class ltt_liable(Variable):
    label = "Liable for Land Transaction Tax"
    documentation = (
        "Whether the household is liable to pay the Wales Land Transaction Tax"
    )
    entity = Household
    definition_period = YEAR
    value_type = bool

    def formula(household, period):
        country = household("country", period)
        return country == country.possible_values.WALES


class ltt_on_residential_property_transactions(Variable):
    label = "LTT on residential property"
    documentation = (
        "Land Transaction Tax charge on residential property transactions"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        ltt = parameters(period).gov.wra.land_transaction_tax
        main_home_price = household(
            "main_residential_property_purchased", period
        )
        primary_residential_purchase_tax = ltt.residential.primary.calc(
            main_home_price
        )
        second_home_price = household(
            "additional_residential_property_purchased", period
        )
        additional_residential_purchase_tax = ltt.residential.higher_rate.calc(
            second_home_price
        )
        return (
            primary_residential_purchase_tax
            + additional_residential_purchase_tax
        )


class ltt_on_residential_property_rent(Variable):
    label = "LTT on residential property rent"
    documentation = (
        "Land Transaction Tax charge on residential property rental agreements"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        ltt = parameters(period).gov.wra.land_transaction_tax
        cumulative_rent = household("cumulative_residential_rent", period)
        rent = household("rent", period)
        ltt_on_cumulative_rent = ltt.rent.calc(cumulative_rent)
        ltt_on_total_rent = ltt.rent.calc(cumulative_rent + rent)
        return ltt_on_total_rent - ltt_on_cumulative_rent


class ltt_on_non_residential_property_transactions(Variable):
    label = "LTT on non-residential property transactions"
    documentation = (
        "Land Transaction Tax charge on non-residential property transactions"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        ltt = parameters(period).gov.wra.land_transaction_tax
        price = household("non_residential_property_purchased", period)
        non_residential_purchase_tax = ltt.non_residential.calc(price)
        return household("ltt_liable", period) * non_residential_purchase_tax


class ltt_on_non_residential_property_rent(Variable):
    label = "LTT on non-residential property rent"
    documentation = "Land Transaction Tax charge on non-residential property rental agreements"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        ltt = parameters(period).gov.wra.land_transaction_tax
        cumulative_rent = household("cumulative_non_residential_rent", period)
        rent = household("non_residential_rent", period)
        ltt_on_cumulative_rent = ltt.rent.calc(cumulative_rent)
        ltt_on_total_rent = ltt.rent.calc(cumulative_rent + rent)
        return ltt_on_total_rent - ltt_on_cumulative_rent


class ltt_on_transactions(Variable):
    label = "LTT on property transactions"
    documentation = "Land Transaction Tax on property transfers"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period):
        LTT_TRANSACTION_VARIABLES = [
            "ltt_on_residential_property_transactions",
            "ltt_on_non_residential_property_transactions",
        ]
        return add(household, period, LTT_TRANSACTION_VARIABLES)


class ltt_on_rent(Variable):
    label = "LTT on property rental"
    documentation = "Land Transaction Tax on property rental agreements"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period):
        LTT_RENT_VARIABLES = [
            "ltt_on_residential_property_rent",
            "ltt_on_non_residential_property_rent",
        ]
        return add(household, period, LTT_RENT_VARIABLES)


class land_transaction_tax(Variable):
    label = "Land Transaction Tax"
    documentation = "Total tax liability for Land Transaction Tax"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period):
        ltt_if_liable = add(
            household, period, ["ltt_on_transactions", "ltt_on_rent"]
        )
        return household("ltt_liable", period) * ltt_if_liable


class expected_ltt(Variable):
    label = "Land Transaction Tax (expected)"
    documentation = "Expected value of Land Transaction Tax"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period):
        property_sale_rate = household.state("property_sale_rate", period)
        land_transaction_tax = household("land_transaction_tax", period)
        return property_sale_rate * land_transaction_tax


class baseline_expected_ltt(Variable):
    label = "Land Transaction Tax (baseline, expected)"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP


class change_in_expected_ltt(Variable):
    label = "average per-year LTT"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    formula = change_over_baseline(expected_ltt)



import streamlit as st
import pandas as pd
import plotly.express as px
import time
from policyengine_uk import REPO

st.header("OpenFisca-UK calibration dashboard")
st.subheader("Inspect training progress by metric and year")

df = pd.read_csv(REPO / "data/storage/training_log_run_1.csv")
df["name_without_year"] = df["name"].apply(
    lambda x: ".".join(x.split(".")[:-1])
)
df["year"] = df.name.str.split(".").str[-1]
df["Predicted"] = df.pred
df["Actual"] = df.actual
df["Error"] = df.pred - df.actual
df["Absolute error"] = (df.pred - df.actual).abs()
df["Relative error"] = (df.pred - df.actual) / df.actual
df["Absolute relative error"] = (df.pred - df.actual).abs() / df.pred
df["Epoch"] = df.epoch

metric = st.selectbox("Metric", df.name.str[:-5].unique())
rerun_every = st.selectbox(
    "Rerun every", ["15 seconds", "minute", "3 minutes", "don't rerun"]
)

subset_df = df[df.name_without_year == metric]


fig_2 = px.line(
    subset_df,
    y="Absolute relative error",
    x="Epoch",
    animation_frame="year",
).update_layout(
    title="Relative error by epoch",
    yaxis_tickformat=".1%",
)

st.write(fig_2)

fig = px.line(
    subset_df,
    y=["Predicted", "Actual"],
    x="Epoch",
    animation_frame="year",
).update_layout(
    title="Outputs by epoch",
)

st.write(fig)

if rerun_every == "15 seconds":
    time.sleep(15)
    st.experimental_rerun()
elif rerun_every == "minute":
    time.sleep(60)
    st.experimental_rerun()
elif rerun_every == "3 minutes":
    time.sleep(180)
    st.experimental_rerun()
else:
    pass



from policyengine_uk.model_api import *


class residential_property_value(Variable):
    label = "Residential property value"
    documentation = "Total value of all owned residential property"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    quantity_type = STOCK

    def formula(household, period, parameters):
        return household("main_residence_value", period) + household(
            "other_residential_property_value", period
        )


class property_wealth(Variable):
    label = "Property wealth"
    documentation = "Total property wealth across all owned properties"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    quantity_type = STOCK

    def formula(household, period):
        VARIABLES = [
            "residential_property_value",
            "non_residential_property_value",
        ]
        return add(household, period, VARIABLES)



from policyengine_uk.model_api import *


class ebr_energy_bills_credit(Variable):
    label = "Energy bills credit (EBR)"
    documentation = "Credit on energy bills under the Energy Bills Rebate scheme. Modeled as a flat transfer."
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        ebr = parameters(period).gov.treasury.energy_bills_rebate
        return ebr.energy_bills_credit



import numpy as np
from policyengine_core.parameters import ParameterNode, Parameter
from numpy.typing import ArrayLike
from itertools import chain
from functools import reduce
from typing import Callable, Iterable, List, Tuple, Type
from policyengine_uk import Microsimulation
from policyengine_uk.parameter_tree import parameters
import tensorflow as tf


@staticmethod
def weighted_squared_relative_deviation(
    pred: tf.Tensor, actual: ArrayLike
) -> tf.Tensor:
    """Computes the squared relative deviation between two tensors, weighted by the actual value.

    Args:
        pred (tf.Tensor): Predicted values.
        actual (ArrayLike): Actual values.

    Returns:
        tf.Tensor: The weighted squared relative deviation.
    """
    return (pred / (actual + 1) - 1) ** 2


class LossCategory:
    weight: float = 1
    label: str
    category: str
    parameter_folder: ParameterNode
    comparison_loss_function: Callable = weighted_squared_relative_deviation
    initial_train_loss: float = None
    initial_val_loss: float = None
    cache = {}
    use_cache = False

    def __init__(
        self, years: List[int], year: int, weight: float, sim: Microsimulation
    ):
        self.years = years or self.years
        self.year = year
        self.weight = weight
        self.sim = sim
        self.calibration_parameters = parameters.calibration(
            f"{self.year}-01-01"
        )
        self.initialise()

    def initialise(self):
        pass

    def get_loss_subcomponents(
        household_weights: tf.Tensor,
    ) -> Iterable[Tuple]:
        """Gathers predictions used to measure weight performance.

        Args:
            sim (Microsimulation): A microsimulation from which to draw demographic data.
            household_weights (tf.Tensor): Household weights for each year.
            year (int): The year to test on.

        Returns:
            Iterable[Tuple]: A list of (name, pred, actual) triplets.
        """
        raise NotImplementedError()

    def compute(
        self,
        sim: Microsimulation,
        household_weights: tf.Tensor,
        excluded_metrics: List[str],
        validation: bool = False,
    ) -> Tuple[tf.Tensor, dict]:
        loss = tf.constant(0, dtype=tf.float32)
        log = []
        for name, pred, actual in self.get_loss_subcomponents(
            household_weights[self.year - self.years[0]],
        ):
            name += f".{self.year}"
            if name not in excluded_metrics:
                l = self.comparison_loss_function(pred, actual)
                log += [
                    dict(
                        name=name,
                        pred=float(pred.numpy()),
                        actual=float(actual),
                        loss=float(l.numpy()),
                        category=self.category,
                    )
                ]
                loss += l
        if not validation and self.initial_train_loss is None:
            self.initial_train_loss = loss.numpy()
        elif validation and self.initial_val_loss is None:
            self.initial_val_loss = loss.numpy()
        initial_value = (
            self.initial_val_loss if validation else self.initial_train_loss
        )
        if initial_value == 0:
            initial_value += 1
        for entry in log:
            entry["loss"] /= initial_value
            entry["loss"] *= self.weight
        if len(log) == 0:
            return loss, log
        return loss / initial_value * self.weight, log

    def get_metrics(self) -> List[Parameter]:
        return []
        return list(
            filter(
                lambda p: isinstance(p, Parameter),
                self.parameter_folder.get_descendants(),
            )
        )

    def get_metric_names(self) -> List[str]:
        return []
        for parameter in self.parameter_folder.get_descendants():
            if isinstance(parameter, Parameter):
                yield f"{parameter.name}.{self.year}"



from policyengine_uk.model_api import *


class pip(Variable):
    label = "Personal Independence Payment"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP

    formula = sum_of_variables(["pip_dl", "pip_m"])


class PIPCategory(Enum):
    STANDARD = "Standard"
    ENHANCED = "Enhanced"
    NONE = "None"



from policyengine_uk.model_api import *


class lbtt_on_residential_property_transactions(Variable):
    label = "LBTT on residential property"
    documentation = "LBTT charge on purchase of residential property"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        lbtt = parameters(period).gov.revenue_scotland.lbtt
        # Tax on main-home purchases
        price = household("main_residential_property_purchased", period)
        residential_purchase_qualifies_as_first_buy = household(
            "main_residential_property_purchased_is_first_home", period
        )
        main_residential_purchase_tax = where(
            residential_purchase_qualifies_as_first_buy,
            lbtt.residential.first_time_buyer_rate.calc(price),
            lbtt.residential.rate.calc(price),
        )
        # Tax on second-home purchases
        second_home_price = household(
            "additional_residential_property_purchased", period
        )
        lbtt2 = lbtt.residential.rate.calc(second_home_price)
        surcharge = (
            lbtt.residential.additional_residence_surcharge * second_home_price
        )
        additional_residential_purchase_tax = lbtt2 + surcharge
        return (
            main_residential_purchase_tax + additional_residential_purchase_tax
        )


class lbtt_on_residential_property_rent(Variable):
    label = "LBTT on residential property rent"
    documentation = "LBTT charge on rental of residential property"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        lbtt = parameters(period).gov.revenue_scotland.lbtt
        cumulative_rent = household("cumulative_residential_rent", period)
        rent = household("rent", period)
        lbtt_cumulative_rent = lbtt.rent.calc(cumulative_rent)
        lbtt_total_rent = lbtt.rent.calc(cumulative_rent + rent)
        return lbtt_total_rent - lbtt_cumulative_rent


class lbtt_on_non_residential_property_transactions(Variable):
    label = "LBTT on non-residential property transactions"
    documentation = "LBTT charge from purchase of non-residential property"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        lbtt = parameters(period).gov.revenue_scotland.lbtt
        price = household("non_residential_property_purchased", period)
        return lbtt.non_residential.calc(price)


class lbtt_on_non_residential_property_rent(Variable):
    label = "LBTT on non-residential property"
    documentation = (
        "LBTT charge from purchase or rental of non-residential property"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        lbtt = parameters(period).gov.revenue_scotland.lbtt
        cumulative_rent = household("cumulative_non_residential_rent", period)
        rent = household("non_residential_rent", period)
        lbtt_cumulative_rent = lbtt.rent.calc(cumulative_rent)
        lbtt_total_rent = lbtt.rent.calc(cumulative_rent + rent)
        return lbtt_total_rent - lbtt_cumulative_rent


class lbtt_liable(Variable):
    label = "Liable for Land and Buildings Transaction Tax"
    documentation = "Whether the household is liable for Land and Buildings Transaction Tax"
    entity = Household
    definition_period = YEAR
    value_type = bool
    unit = GBP

    def formula(household, period):
        country = household("country", period)
        countries = country.possible_values
        return country == countries.SCOTLAND


class lbtt_on_transactions(Variable):
    label = "LBTT on property transactions"
    documentation = "Land and Buildings Transaction Tax on property transfers"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period):
        LBTTS = [
            "lbtt_on_residential_property_transactions",
            "lbtt_on_non_residential_property_transactions",
        ]
        return add(household, period, LBTTS)


class lbtt_on_rent(Variable):
    label = "LBTT on property rental"
    documentation = (
        "Land and Buildings Transaction Tax on property rental agreements"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period):
        LBTTS = [
            "lbtt_on_residential_property_rent",
            "lbtt_on_non_residential_property_rent",
        ]
        return add(household, period, LBTTS)


class land_and_buildings_transaction_tax(Variable):
    label = "Land and Buildings Transaction Tax"
    documentation = "Total tax liability for Scotland's LBTT"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period):
        LBTTS = [
            "lbtt_on_transactions",
            "lbtt_on_rent",
        ]
        lbtt_if_liable = add(household, period, LBTTS)
        return household("lbtt_liable", period) * lbtt_if_liable


class expected_lbtt(Variable):
    label = "Land and Buildings Transaction Tax (expected)"
    documentation = "Expected value of LBTT"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period):
        property_sale_rate = household.state("property_sale_rate", period)
        lbtt = household("land_and_buildings_transaction_tax", period)
        return property_sale_rate * lbtt


class baseline_expected_lbtt(Variable):
    label = "LBTT (expected, baseline)"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP


class change_in_expected_lbtt(Variable):
    label = "average per-year LBTT"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    formula = change_over_baseline(expected_lbtt)



from policyengine_uk.model_api import *

label = "Income"
description = "Financial income received by individuals."


class employment_income(Variable):
    value_type = float
    entity = Person
    label = "employment income"
    documentation = (
        "Total income from employment. Include wages, bonuses, tips, etc."
    )
    definition_period = YEAR
    unit = GBP
    reference = "Income Tax (Earnings and Pensions) Act 2003 s. 1(1)(a)"
    quantity_type = FLOW
    category = INCOME
    uprating = "calibration.programs.employment_income.budgetary_impact.UNITED_KINGDOM"


class pension_income(Variable):
    value_type = float
    entity = Person
    label = "pension income"
    documentation = "Income from private or occupational pensions (not including the State Pension)"
    definition_period = YEAR
    unit = GBP
    reference = "Income Tax (Earnings and Pensions) Act 2003 s. 1(1)(b)"
    quantity_type = FLOW
    uprating = "calibration.programs.pension_income.budgetary_impact.ENGLAND"


class state_pension(Variable):
    value_type = float
    entity = Person
    label = "State Pension"
    definition_period = YEAR
    unit = GBP
    documentation = "Gross State Pension payments"
    quantity_type = FLOW

    def formula(person, period, parameters):
        return person("state_pension_reported", period)


class self_employment_income(Variable):
    value_type = float
    entity = Person
    label = "self-employment income"
    documentation = "Income from self-employment profits"
    definition_period = YEAR
    unit = GBP
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 1(1)(a)"
    quantity_type = FLOW
    uprating = "calibration.programs.self_employment_income.budgetary_impact.UNITED_KINGDOM"


class property_income(Variable):
    value_type = float
    entity = Person
    label = "rental income"
    documentation = "Income from rental of property"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 1(1)(b)"
    unit = GBP
    quantity_type = FLOW
    uprating = (
        "calibration.programs.property_income.budgetary_impact.UNITED_KINGDOM"
    )


class savings_interest_income(Variable):
    value_type = float
    entity = Person
    label = "savings interest income"
    documentation = "Income from interest on savings, gross of tax"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 365(1)(a)"
    unit = GBP
    quantity_type = FLOW
    uprating = "calibration.programs.savings_interest_income.budgetary_impact.UNITED_KINGDOM"


class dividend_income(Variable):
    value_type = float
    entity = Person
    label = "dividend income"
    documentation = "Total income from dividends, gross of tax"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 365(1)(b-d)"
    unit = GBP
    quantity_type = FLOW
    uprating = (
        "calibration.programs.dividend_income.budgetary_impact.UNITED_KINGDOM"
    )


class sublet_income(Variable):
    value_type = float
    entity = Person
    label = "sublet income"
    documentation = "Income from subletting properties"
    definition_period = YEAR
    unit = GBP
    uprating = (
        "calibration.programs.property_income.budgetary_impact.UNITED_KINGDOM"
    )


class miscellaneous_income(Variable):
    value_type = float
    entity = Person
    label = "miscellaneous income"
    documentation = "Income from any other source"
    definition_period = YEAR
    unit = GBP
    uprating = "calibration.programs.employment_income.budgetary_impact.UNITED_KINGDOM"


class private_transfer_income(Variable):
    value_type = float
    entity = Person
    label = "private transfer income"
    documentation = "Income from private transfers"
    definition_period = YEAR
    unit = GBP


class lump_sum_income(Variable):
    value_type = float
    entity = Person
    label = "lump sum income"
    documentation = "Income from lump sums"
    definition_period = YEAR
    unit = GBP


class maintenance_income(Variable):
    value_type = float
    entity = Person
    label = "maintenance payment income"
    documentation = "Income from maintenance payments to you"
    definition_period = YEAR
    unit = GBP



from pathlib import Path
from policyengine_uk.entities import entities
from policyengine_core.taxbenefitsystems import TaxBenefitSystem
from policyengine_core.simulations import (
    Simulation as CoreSimulation,
    Microsimulation as CoreMicrosimulation,
    IndividualSim as CoreIndividualSim,
)
from policyengine_uk.data.datasets.frs.enhanced.stages.imputation.enhanced_frs import (
    EnhancedFRS,
)
from policyengine_uk.parameters.gov.ofgem.price_cap.add_price_cap_parameters import (
    add_price_cap_parameters,
)

COUNTRY_DIR = Path(__file__).parent


class CountryTaxBenefitSystem(TaxBenefitSystem):
    parameters_dir = COUNTRY_DIR / "parameters"
    variables_dir = COUNTRY_DIR / "variables"
    auto_carry_over_input_variables = True
    basic_inputs = [
        "BRMA",
        "local_authority",
        "region",
        "employment_income",
        "age",
    ]

    def __init__(self):
        super().__init__(entities)

        self.parameters = add_price_cap_parameters(
            self.parameters,
            start_year=2019,
            start_quarter=1,
            end_year=2027,
            end_quarter=4,
        )

        self.parameters.add_child("baseline", self.parameters.clone())


system = CountryTaxBenefitSystem()

parameters = system.parameters
variables = system.variables


class Simulation(CoreSimulation):
    default_tax_benefit_system = CountryTaxBenefitSystem
    default_tax_benefit_system_instance = system
    default_calculation_period = 2022
    default_input_period = 2022
    default_role = "member"
    max_spiral_loops = 10


class Microsimulation(CoreMicrosimulation):
    default_tax_benefit_system = CountryTaxBenefitSystem
    default_tax_benefit_system_instance = system
    default_dataset = EnhancedFRS
    default_dataset_year = 2022
    default_calculation_period = 2022
    default_input_period = 2022
    default_role = "member"
    max_spiral_loops = 10


class IndividualSim(CoreIndividualSim):  # Deprecated
    tax_benefit_system = CountryTaxBenefitSystem
    entities = {entity.key: entity for entity in entities}
    default_dataset = EnhancedFRS
    required_entities = None


BASELINE_VARIABLES = {
    variable.name: variable for variable in system.variables.values()
}



import policyengine_uk
from policyengine_uk.model_api import *
from typing import Union


def set_parameter(
    path: Union[Parameter, str], value: float, period: str = "year:2015:10"
) -> Reform:
    if isinstance(path, Parameter):
        path = path.name

    def modifier(parameters: ParameterNode):
        node = parameters
        for name in path.split("."):
            try:
                if "[" not in name:
                    node = node.children[name]
                else:
                    try:
                        name, index = name.split("[")
                        index = int(index[:-1])
                        node = node.children[name].brackets[index]
                    except:
                        raise ValueError(
                            "Invalid bracket syntax (should be e.g. tax.brackets[3].rate"
                        )
            except:
                raise ValueError(
                    f"Could not find the parameter (failed at {name})."
                )
        node.update(period=period, value=value)
        return parameters

    class reform(Reform):
        def apply(self):
            self.modify_parameters(modifier)

    return reform


def change_parameter(
    path: Parameter, value: float, period: str = "year:2015:10"
) -> Reform:
    path = path.name

    def modifier(parameters: ParameterNode):
        node = parameters
        for name in path.split("."):
            try:
                if "[" not in name:
                    node = node.children[name]
                else:
                    try:
                        name, index = name.split("[")
                        index = int(index[:-1])
                        node = node.children[name].brackets[index]
                    except:
                        raise ValueError(
                            "Invalid bracket syntax (should be e.g. tax.brackets[3].rate"
                        )
            except:
                raise ValueError(
                    f"Could not find the parameter (failed at {name})."
                )
        node.update(period=period, value=node.get_at_instant(period))
        return parameters

    class reform(Reform):
        def apply(self):
            self.modify_parameters(modifier)

    return reform



from policyengine_uk.model_api import *


class would_evade_tv_licence_fee(Variable):
    label = "Would evade TV licence fee"
    documentation = (
        "Whether this household would unlawfully evade the TV licence fee."
    )
    entity = Household
    definition_period = YEAR
    value_type = bool

    def formula(household, period, parameters):
        evasion_rate = parameters(period).gov.dcms.bbc.tv_licence.evasion_rate
        return random(household) <= evasion_rate



from policyengine_uk.model_api import *


class petrol_litres(Variable):
    label = "Petrol volume"
    documentation = "Total litres of petrol bought"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        return household("petrol_spending", period) / household(
            "petrol_price", period
        )


class diesel_litres(Variable):
    label = "Diesel volume"
    documentation = "Total litres of diesel bought"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        return household("diesel_spending", period) / household(
            "diesel_price", period
        )


class petrol_price(Variable):
    label = "Price of petrol per litre"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        return parameters(period).household.consumption.fuel.prices.petrol


class diesel_price(Variable):
    label = "Price of diesel per litre"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        return parameters(period).household.consumption.fuel.prices.petrol



from policyengine_uk.model_api import *
from policyengine_uk.variables.misc.categories.lower_middle_or_higher import (
    LowerMiddleOrHigher,
)


class DLA_SC_reported(Variable):
    value_type = float
    entity = Person
    label = "DLA (self-care) (reported)"
    definition_period = YEAR
    unit = GBP


class dla_sc(Variable):
    label = "DLA (self-care)"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(person, period, parameters):
        dla_sc = parameters(period).gov.dwp.dla.self_care
        category = person("dla_sc_category", period)
        return (
            select(
                [
                    category == LowerMiddleOrHigher.HIGHER,
                    category == LowerMiddleOrHigher.MIDDLE,
                    category == LowerMiddleOrHigher.LOWER,
                    category == LowerMiddleOrHigher.NONE,
                ],
                [
                    dla_sc.higher,
                    dla_sc.middle,
                    dla_sc.lower,
                    0,
                ],
            )
            * WEEKS_IN_YEAR
        )


class dla_sc_middle_plus(Variable):
    value_type = bool
    entity = Person
    label = "Receives at least middle-rate DLA (self-care)"
    definition_period = YEAR

    def formula(person, period, parameters):
        return is_in(
            person("dla_sc_category", period),
            [LowerMiddleOrHigher.MIDDLE, LowerMiddleOrHigher.HIGHER],
        )


class receives_highest_dla_sc(Variable):
    label = "Receives the highest DLA (self-care) category"
    entity = Person
    definition_period = YEAR
    value_type = bool

    def formula(person, period, parameters):
        return person("dla_sc_category", period) == LowerMiddleOrHigher.HIGHER



from policyengine_uk.model_api import *
from policyengine_uk.variables.gov.dwp.pension_credit.pension_credit import (
    pension_credit,
)


class would_claim_pc(Variable):
    label = "Would claim Pension Credit"
    entity = BenUnit
    definition_period = YEAR
    value_type = bool

    def formula(benunit, period, parameters):
        reported_pc = add(benunit, period, ["pension_credit_reported"]) > 0
        claims_all_entitled_benefits = benunit(
            "claims_all_entitled_benefits", period
        )
        baseline = benunit("baseline_pension_credit_entitlement", period) > 0
        eligible = benunit("pension_credit_entitlement", period) > 0
        takeup_rate = parameters(period).gov.dwp.pension_credit.takeup
        return select(
            [
                reported_pc | claims_all_entitled_benefits,
                ~baseline & eligible,
                True,
            ],
            [
                True,
                random(benunit) < takeup_rate,
                False,
            ],
        )



from ...loss_category import LossCategory
from policyengine_uk.tools.simulation import Microsimulation
import tensorflow as tf
import numpy as np
from policyengine_core.parameters import ParameterNode, Parameter
from typing import Iterable, Tuple
from policyengine_uk.parameter_tree import parameters


class Populations(LossCategory):
    name = "Populations"
    category = "Populations"

    def initialise(self):
        self.people = self.sim.calc(
            "people", map_to="household", period=self.year
        ).values
        self.comparisons = []
        age = self.sim.calc("age").values
        sex = self.sim.calc("gender").values
        region = self.sim.calc("region", map_to="person").values
        age_sex_region = (
            self.calibration_parameters.demographics.populations.by_age_sex_region
        )
        total_population = 0
        for lower_age in list(range(0, 80, 10)) + [79]:
            meets_age_criteria = (age >= lower_age) & (age < lower_age + 10)
            age_population = 0
            age_string = (
                f"BETWEEN_{lower_age}_{lower_age + 10}"
                if lower_age < 79
                else "OVER_80"
            )
            for target_sex in ("MALE", "FEMALE"):
                age_sex_population = 0
                for target_region in self.sim.calc("region").unique():
                    meets_all_criteria = (
                        meets_age_criteria
                        & (region == target_region)
                        & (sex == target_sex)
                    )
                    actual_population = (
                        age_sex_region._children[target_sex]
                        ._children[target_region]
                        ._children[age_string]
                    )
                    self.comparisons += [
                        (
                            f"{target_region}_{target_sex}_{lower_age}_{lower_age + 10}",
                            self.sim.map_result(
                                meets_all_criteria, "person", "household"
                            ),
                            actual_population,
                        )
                    ]
                    age_sex_population += actual_population

                self.comparisons += [
                    (
                        f"{target_sex}_{lower_age}_{lower_age + 10}",
                        self.sim.map_result(
                            meets_age_criteria & (sex == target_sex),
                            "person",
                            "household",
                        ),
                        age_sex_population,
                    )
                ]

                age_population += age_sex_population

            self.comparisons += [
                (
                    f"{lower_age}_{lower_age + 10}",
                    self.sim.map_result(
                        meets_age_criteria, "person", "household"
                    ),
                    age_population,
                )
            ]

            total_population += age_population

        self.comparisons += [
            (
                f"people",
                self.sim.map_result(np.ones_like(age), "person", "household"),
                total_population,
            )
        ]

    def get_loss_subcomponents(
        self, household_weights: tf.Tensor
    ) -> Iterable[Tuple]:
        for name, values, target in self.comparisons:
            yield name, tf.reduce_sum(values * household_weights), target

    def get_metric_names(self) -> Iterable[str]:
        return [x[0] for x in self.comparisons]



from policyengine_uk.model_api import *
import warnings


class bi_individual_phaseout(Variable):
    label = "Basic income phase-out (individual)"
    documentation = (
        "Reduction in basic income from individual-level phase-outs."
    )
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(person, period, parameters):
        income = person("total_income", period)
        bi = parameters(period).gov.contrib.ubi_center.basic_income
        max_bi = person("bi_maximum", period)
        income_over_threshold = max_(
            income - bi.phase_out.individual.threshold, 0
        )
        uncapped_deduction = (
            bi.phase_out.individual.rate * income_over_threshold
        )
        return min_(max_bi, uncapped_deduction)


class bi_household_phaseout(Variable):
    label = "Basic income phase-out (household)"
    documentation = (
        "Reduction in basic income from household-level phase-outs."
    )
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(person, period, parameters):
        income = person("total_income", period)
        household = person.household
        household_income = household.sum(income)
        bi = parameters(period).gov.contrib.ubi_center.basic_income
        remaining_bi = person("bi_maximum", period) - person(
            "bi_individual_phaseout", period
        )  # Basic income remaining after individual-level phaseouts

        household_bi = household.sum(remaining_bi)
        income_over_threshold = max_(
            household_income - bi.phase_out.household.threshold,
            0,
        )
        uncapped_deduction = (
            bi.phase_out.household.rate * income_over_threshold
        )
        capped_deduction = min_(household_bi, uncapped_deduction)

        warnings.filterwarnings("ignore")

        percent_reduction = where(
            household_bi > 0, capped_deduction / household_bi, 0
        )
        return percent_reduction * remaining_bi


class bi_phaseout(Variable):
    label = "Basic income phase-out"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    formula = sum_of_variables(
        ["bi_individual_phaseout", "bi_household_phaseout"]
    )



from policyengine_uk.model_api import *


class JSA_income_reported(Variable):
    value_type = float
    entity = Person
    label = "JSA (income-based) (reported amount)"
    definition_period = YEAR
    unit = GBP


class JSA_income_eligible(Variable):
    value_type = bool
    entity = BenUnit
    label = "Eligibility for income-based JSA"
    documentation = "Whether the benefit unit is eligible for income-based Jobseekers' Allowance"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        # Calculate work hours eligibility.
        hours_limit = parameters(period).gov.dwp.JSA.hours
        hours = benunit("benunit_weekly_hours", period)
        single = benunit("is_single", period)
        hours_eligible_as_single = single & (hours < hours_limit.single)
        couple = benunit("is_couple", period)
        hours_eligible_as_couple = couple & (hours < hours_limit.couple)
        hours_eligible = hours_eligible_as_single | hours_eligible_as_couple
        # Benefit units with state pension age people are ineligible.
        all_under_SP_age = ~benunit.any(benunit.members("is_SP_age", period))
        # Must have at least one unemployed person.
        employment_statuses = benunit.members("employment_status", period)
        unemployed_members = (
            employment_statuses
            == employment_statuses.possible_values.UNEMPLOYED
        )
        any_unemployed = benunit.any(unemployed_members)
        # Cannot claim Income Support.
        not_on_income_support = benunit("income_support", period) == 0
        already_claiming = aggr(benunit, period, ["JSA_income_reported"]) > 0
        return (
            hours_eligible
            & all_under_SP_age
            & any_unemployed
            & not_on_income_support
            & already_claiming
        )


class JSA_income_applicable_amount(Variable):
    value_type = float
    entity = BenUnit
    label = "Maximum amount of JSA (income-based)"
    documentation = "Maximum amount of income-based Jobseeker's Allowance"
    definition_period = YEAR
    reference = "Jobseekers Act 1995 s. 4"
    unit = GBP

    def formula(benunit, period, parameters):
        income = parameters(period).gov.dwp.JSA.income
        age = benunit("youngest_adult_age", period)
        single = benunit("is_single", period)
        couple = benunit("is_couple", period)
        single_18_24 = single & (age < 25)
        single_over_25 = single & (age >= 25)
        pa_single_18_24 = single_18_24 * income.amount_18_24
        pa_single_over_25 = single_over_25 * income.amount_over_25
        pa_couple = couple * income.couple
        weekly_personal_allowance = (
            pa_single_18_24 + pa_single_over_25 + pa_couple
        )
        personal_allowance = weekly_personal_allowance * WEEKS_IN_YEAR
        premiums = benunit("benefits_premiums", period)
        amount_if_claims = personal_allowance + premiums
        eligible = benunit("JSA_income_eligible", period)
        claims = benunit("would_claim_JSA", period)
        return amount_if_claims * eligible * claims


class would_claim_JSA(Variable):
    value_type = bool
    entity = BenUnit
    label = "Would claim income-based JSA"
    documentation = (
        "Whether this family would claim income-based JSA if eligible"
    )
    definition_period = YEAR

    def formula(benunit, period, parameters):
        reported_JSA = aggr(benunit, period, ["JSA_income_reported"]) > 0
        return reported_JSA


class JSA_income_applicable_income(Variable):
    value_type = float
    entity = BenUnit
    label = "Relevant income for JSA (income-based) means test"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        JSA = parameters(period).gov.dwp.JSA
        INCOME_COMPONENTS = [
            "employment_income",
            "self_employment_income",
            "property_income",
            "pension_income",
        ]
        bi = parameters(period).gov.contrib.ubi_center.basic_income
        if bi.interactions.include_in_means_tests:
            INCOME_COMPONENTS.append("basic_income")
        income = aggr(benunit, period, INCOME_COMPONENTS)
        tax = aggr(
            benunit,
            period,
            ["income_tax", "national_insurance"],
        )
        income += aggr(benunit, period, ["social_security_income"])
        income -= tax
        income -= aggr(benunit, period, ["pension_contributions"]) * 0.5
        # Calculate disregard.
        family_type = benunit("family_type", period)
        families = family_type.possible_values
        single = family_type == families.SINGLE
        single_disregard = single * JSA.income.income_disregard_single
        couple = benunit("is_couple", period)
        couple_disregard = couple * JSA.income.income_disregard_couple
        lone_parent = family_type == families.LONE_PARENT
        lone_parent_disregard = (
            lone_parent * JSA.income.income_disregard_lone_parent
        )
        weekly_disregard = (
            single_disregard + couple_disregard + lone_parent_disregard
        )
        disregard = weekly_disregard * WEEKS_IN_YEAR
        return max_(0, income - disregard)


class JSA_income(Variable):
    value_type = float
    entity = BenUnit
    label = "JSA (income-based)"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        applicable_amount = benunit("JSA_income_applicable_amount", period)
        applicable_income = benunit("JSA_income_applicable_income", period)
        amount_if_claims = max_(0, applicable_amount - applicable_income)
        claims = benunit("would_claim_JSA", period)
        eligible = benunit("JSA_income_eligible", period)
        return amount_if_claims * claims * eligible


class JSA(Variable):
    value_type = float
    entity = BenUnit
    label = "Amount of Jobseeker's Allowance for this family"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        JSA_contrib = aggr(benunit, period, ["JSA_contrib"])
        JSA_income = benunit("JSA_income", period)
        return JSA_contrib + JSA_income



from policyengine_uk.model_api import *

"""
The calculation of income tax is specified by the Income Tax Act 2007 s. 23
which outlines the main steps. Step 1 indicates that the components of 'total income'
should be added together, but the Act itself does not specify what those components are.
Instead, other Acts (the Income Tax (Earnings and Pensions) Act 2003 and the Income Tax 
(Trading and Other Income) Act 2005) impose the charges to income tax on each component of
income. The relevant sections of each act imposing charges to income tax are given 
for each component of income below. The components are guided by the article at 
https://www.pruadviser.co.uk/knowledge-literature/knowledge-library/the-seven-steps-required-to-calculate-an-individuals-income-tax-liability/
"""


class total_pension_income(Variable):
    label = "Total pension income"
    documentation = "Private, personal and State Pension income"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP

    formula = sum_of_variables(["pension_income", "state_pension"])


class social_security_income(Variable):
    value_type = float
    entity = Person
    label = "Income from social security for tax purposes"
    definition_period = YEAR
    reference = "Income Tax (Earnings and Pensions) Act 2003 s. 1(1)(b)"
    unit = GBP

    def formula(person, period, parameters):
        COMPONENTS = [
            "state_pension",
            "incapacity_benefit",
            "JSA_contrib",
            "ESA_contrib",
            "carers_allowance",
        ]
        return add(person, period, COMPONENTS)



from policyengine_uk.model_api import *


class wealth_tax(Variable):
    label = "Wealth tax"
    documentation = "Annual tax on household net wealth"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        wealth = household("total_wealth", period)
        tax = parameters(period).gov.contrib.ubi_center.wealth_tax
        return tax.calc(wealth)



def test_synth_runs():
    from policyengine_uk import Microsimulation
    from policyengine_uk.data import SynthFRS

    if 2022 not in SynthFRS.years:
        SynthFRS.download(2022)

    sim = Microsimulation(dataset=SynthFRS)
    sim.calculate("household_net_income")



from policyengine_uk.model_api import *


class standard_minimum_guarantee(Variable):
    label = "Standard Minimum Guarantee"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/uksi/2002/1792/regulation/6"

    def formula(benunit, period, parameters):
        relation_type = benunit("relation_type", period)
        pc = parameters(period).gov.dwp.pension_credit
        weekly_rate = pc.guarantee_credit.minimum_guarantee[relation_type]
        return weekly_rate * WEEKS_IN_YEAR



from policyengine_uk.model_api import *


class total_wealth(Variable):
    label = "Total wealth"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    formula = sum_of_variables(["property_wealth", "corporate_wealth"])



from policyengine_core.model_api import *
from policyengine_core import periods


def str_to_instant(s):
    return periods.Instant(tuple(map(lambda s: int(s), s.split("-"))))


def backdate_parameters(
    root: str = None, first_instant: str = "2010-01-01"
) -> Reform:
    first_instant = str_to_instant(first_instant)

    def modifier(parameters):
        node = parameters
        if root is not None:
            for name in root.split("."):
                node = node[name]
        for param in node.get_descendants():
            if hasattr(param, "values_list"):
                earliest = param.values_list[-1]
                earliest_value = earliest.value
                earliest_instant = str_to_instant(earliest.instant_str)
                if first_instant < earliest_instant:
                    num_days = (
                        earliest_instant.date - first_instant.date
                    ).days
                    param.update(
                        period=periods.Period(
                            ("day", first_instant, num_days)
                        ),
                        value=earliest_value,
                    )
        return parameters

    class backdating_reform(Reform):
        def apply(self):
            self.modify_parameters(modifier)

    return backdating_reform



from policyengine_uk.model_api import *


class is_disabled_for_benefits(Variable):
    value_type = bool
    entity = Person
    label = "Has a disability"
    documentation = "Whether this person is disabled for benefits purposes"
    definition_period = YEAR
    reference = "Child Tax Credit Regulations 2002 s. 8"

    def formula(person, period, parameters):
        QUALIFYING_BENEFITS = [
            "dla",
            "pip",
        ]
        return add(person, period, QUALIFYING_BENEFITS) > 0


class is_enhanced_disabled_for_benefits(Variable):
    value_type = bool
    entity = Person
    label = "Whether meets the middle disability benefit entitlement"
    definition_period = YEAR

    def formula(person, period, parameters):
        DLA_requirement = (
            parameters(period).gov.dwp.dla.self_care.higher * WEEKS_IN_YEAR
        )
        return person("dla_sc", period) >= DLA_requirement


class is_severely_disabled_for_benefits(Variable):
    value_type = bool
    entity = Person
    label = "Has a severe disability"
    documentation = (
        "Whether this person is severely disabled for benefits purposes"
    )
    definition_period = YEAR
    reference = "Child Tax Credit Regulations 2002 s. 8"

    def formula(person, period, parameters):
        benefit = parameters(period).gov.dwp
        THRESHOLD_SAFETY_GAP = 10 * WEEKS_IN_YEAR
        paragraph_3 = (
            person("dla_sc", period)
            >= benefit.dla.self_care.higher * WEEKS_IN_YEAR
            - THRESHOLD_SAFETY_GAP
        )
        paragraph_4 = (
            person("pip_dl", period)
            >= benefit.pip.daily_living.enhanced * WEEKS_IN_YEAR
            - THRESHOLD_SAFETY_GAP
        )
        paragraph_5 = person("AFCS", period) > 0
        return sum([paragraph_3, paragraph_4, paragraph_5]) > 0


class num_disabled_children(Variable):
    value_type = int
    entity = BenUnit
    label = "Number of disabled children"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        child = benunit.members("is_child_or_QYP", period)
        disabled = benunit.members("is_disabled_for_benefits", period)
        return benunit.sum(child & disabled)


class num_enhanced_disabled_children(Variable):
    value_type = int
    entity = BenUnit
    label = "Number of enhanced disabled children"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        child = benunit.members("is_child_or_QYP", period)
        enhanced_disabled = benunit.members(
            "is_enhanced_disabled_for_benefits", period
        )
        return benunit.sum(child & enhanced_disabled)


class num_severely_disabled_children(Variable):
    value_type = int
    entity = BenUnit
    label = "Number of severely disabled children"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        child = benunit.members("is_child_or_QYP", period)
        severely_disabled = benunit.members(
            "is_severely_disabled_for_benefits", period
        )
        return benunit.sum(child & severely_disabled)


class num_disabled_adults(Variable):
    value_type = int
    entity = BenUnit
    label = "Number of disabled adults"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        adult = benunit.members("is_adult", period)
        disabled = benunit.members("is_disabled_for_benefits", period)
        return benunit.sum(adult & disabled)


class num_enhanced_disabled_adults(Variable):
    value_type = int
    entity = BenUnit
    label = "Number of enhanced disabled adults"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        adult = benunit.members("is_adult", period)
        enhanced_disabled = benunit.members(
            "is_enhanced_disabled_for_benefits", period
        )
        return benunit.sum(adult & enhanced_disabled)


class num_severely_disabled_adults(Variable):
    value_type = int
    entity = BenUnit
    label = "Number of severely disabled adults"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        adult = benunit.members("is_adult", period)
        severely_disabled = benunit.members(
            "is_severely_disabled_for_benefits", period
        )
        return benunit.sum(adult & severely_disabled)


class disability_premium(Variable):
    value_type = float
    entity = BenUnit
    label = "Disability premium"
    definition_period = YEAR
    reference = "The Social Security Amendment (Enhanced Disability Premium) Regulations 2000"
    unit = GBP

    def formula(benunit, period, parameters):
        dis = parameters(period).gov.dwp.disability_premia
        single = benunit("is_single", period.this_year)
        couple = benunit("is_couple", period.this_year)
        single_premium = single * dis.disability_single
        couple_premium = couple * dis.disability_couple
        has_disabled_adults = (
            benunit("num_disabled_adults", period.this_year) > 0
        )
        weekly_amount = single_premium + couple_premium
        return weekly_amount * WEEKS_IN_YEAR * has_disabled_adults


class severe_disability_premium(Variable):
    value_type = float
    entity = BenUnit
    label = "Severe disability premium"
    definition_period = YEAR
    reference = "The Social Security Amendment (Enhanced Disability Premium) Regulations 2000"
    unit = GBP

    def formula(benunit, period, parameters):
        dis = parameters(period).gov.dwp.disability_premia
        single = benunit("is_single", period.this_year)
        couple = benunit("is_couple", period.this_year)
        single_premium = single * dis.severe_single
        couple_premium = couple * dis.severe_couple
        has_severely_disabled_adults = (
            benunit("num_severely_disabled_adults", period.this_year) > 0
        )
        weekly_amount = single_premium + couple_premium
        return weekly_amount * WEEKS_IN_YEAR * has_severely_disabled_adults


class enhanced_disability_premium(Variable):
    value_type = float
    entity = BenUnit
    label = "Enhanced disability premium"
    definition_period = YEAR
    reference = "The Social Security Amendment (Enhanced Disability Premium) Regulations 2000"
    unit = GBP

    def formula(benunit, period, parameters):
        dis = parameters(period).gov.dwp.disability_premia
        single = benunit("is_single", period.this_year)
        couple = benunit("is_couple", period.this_year)
        single_premium = single * dis.enhanced_single
        couple_premium = couple * dis.enhanced_couple
        has_enhanced_disabled_adults = (
            benunit("num_enhanced_disabled_adults", period.this_year) > 0
        )
        weekly_amount = single_premium + couple_premium
        return weekly_amount * WEEKS_IN_YEAR * has_enhanced_disabled_adults



# -*- coding: utf-8 -*-

# This file defines the entities needed by our legislation.
from policyengine_core.entities import build_entity

State = build_entity(
    key="state",
    plural="states",
    label="State",
    roles=[
        {
            "key": "member",
            "plural": "members",
            "label": "Member",
            "doc": "A person who is a citizen of a country.",
        }
    ],
)

Household = build_entity(
    key="household",
    plural="households",
    label="Household",
    doc="Each household may contain multiple families, or benefit units.",
    roles=[
        {
            "key": "member",
            "plural": "members",
            "label": "Member",
            "doc": "A person who is a citizen of a country.",
        }
    ],
    containing_entities=["state"],
)

BenUnit = build_entity(
    key="benunit",
    plural="benunits",
    label="Benefit unit",
    doc="A family is also known as a benefit unit, and contains parents and dependent children.",
    roles=[
        {
            "key": "member",
            "plural": "members",
            "label": "Member",
            "doc": "A person who is a citizen of a country.",
        }
    ],
    containing_entities=["household", "state"],
)

Person = build_entity(
    key="person",
    plural="people",
    label="Person",
    doc="An individual. The minimal legal entity on which a legislation might be applied.",
    is_person=True,
)

entities = [State, Household, BenUnit, Person]



from policyengine_uk.model_api import *


class cliff_evaluated(Variable):
    value_type = bool
    entity = Person
    label = "cliff evaluated"
    unit = GBP
    documentation = "Whether this person's cliff has been simulated. If not, then the cliff gap is assumed to be zero."
    definition_period = YEAR

    def formula(person, period, parameters):
        adult_index_values = person("adult_index", period)
        cliff_adult_count = 2
        return adult_index_values <= cliff_adult_count


class cliff_gap(Variable):
    value_type = float
    entity = Person
    label = "cliff gap"
    unit = GBP
    documentation = "Amount of income lost if this person's employment income increased by £2,000."
    definition_period = YEAR

    def formula(person, period, parameters):
        DELTA = 2_000
        cliff_values = np.zeros(person.count, dtype=np.float32)
        simulation = person.simulation
        adult_index_values = person("adult_index", period)
        cliff_adult_count = 2
        for adult_index in range(1, 1 + cliff_adult_count):
            alt_simulation = simulation.get_branch(
                f"adult_{adult_index}_2k_pay_rise"
            )
            mask = adult_index_values == adult_index
            for variable in simulation.tax_benefit_system.variables:
                if variable not in simulation.input_variables:
                    alt_simulation.delete_arrays(variable)
            alt_simulation.set_input(
                "employment_income",
                period,
                person("employment_income", period) + mask * DELTA,
            )
            alt_person = alt_simulation.person
            household_net_income = person.household(
                "household_net_income", period
            )
            household_net_income_higher_earnings = alt_person.household(
                "household_net_income", period
            )
            increase = (
                household_net_income_higher_earnings - household_net_income
            )
            cliff_values += where(mask & (increase < 0), -increase, 0)
        return cliff_values


class is_on_cliff(Variable):
    value_type = bool
    entity = Person
    label = "is on a tax-benefit cliff"
    documentation = "Whether this person would be worse off if their employment income were $2,000 higher."
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("cliff_gap", period) > 0



from policyengine_uk.model_api import *
from policyengine_uk.variables.misc.categories.lower_middle_or_higher import (
    LowerMiddleOrHigher,
)
from policyengine_uk.variables.misc.categories.lower_or_higher import (
    LowerOrHigher,
)
from policyengine_uk.variables.gov.dwp.pip.pip import PIPCategory


class dla_sc_category(Variable):
    label = "DLA (Self-care) category"
    documentation = "If you receive the self-care component of Disability Living Allowance, you will be in one of the following categories: Lower, Middle, Higher. If not, select None."
    entity = Person
    definition_period = YEAR
    value_type = Enum
    possible_values = LowerMiddleOrHigher
    default_value = LowerMiddleOrHigher.NONE

    def formula(person, period, parameters):
        dla_sc = parameters(period).gov.dwp.dla.self_care
        SAFETY_MARGIN = 0.1  # Survey reported values could be slightly below eligible values when they should be above due to data manipulation
        reported_weekly_dla_sc = (
            person("DLA_SC_reported", period) / WEEKS_IN_YEAR
        )
        return select(
            [
                reported_weekly_dla_sc >= dla_sc.higher * (1 - SAFETY_MARGIN),
                reported_weekly_dla_sc >= dla_sc.middle * (1 - SAFETY_MARGIN),
                reported_weekly_dla_sc >= dla_sc.lower * (1 - SAFETY_MARGIN),
                True,
            ],
            [
                LowerMiddleOrHigher.HIGHER,
                LowerMiddleOrHigher.MIDDLE,
                LowerMiddleOrHigher.LOWER,
                LowerMiddleOrHigher.NONE,
            ],
        )


class dla_m_category(Variable):
    label = "DLA (mobility) category"
    documentation = "If you receive the mobility component of Disability Living Allowance, you will be in one of the following categories: Lower, Higher. If not, select None."
    entity = Person
    definition_period = YEAR
    value_type = Enum
    possible_values = LowerOrHigher
    default_value = LowerOrHigher.NONE

    def formula(person, period, parameters):
        dla_m = parameters(period).gov.dwp.dla.mobility
        SAFETY_MARGIN = 0.1  # Survey reported values could be slightly below eligible values when they should be above due to data manipulation
        reported_weekly_dla_m = (
            person("DLA_M_reported", period) / WEEKS_IN_YEAR
        )
        return select(
            [
                reported_weekly_dla_m >= dla_m.higher * (1 - SAFETY_MARGIN),
                reported_weekly_dla_m >= dla_m.lower * (1 - SAFETY_MARGIN),
                True,
            ],
            [
                LowerOrHigher.HIGHER,
                LowerOrHigher.LOWER,
                LowerOrHigher.NONE,
            ],
        )


class pip_m_category(Variable):
    label = "PIP (mobility) category"
    documentation = "If you receive the mobility component of the Personal Independence Payment, you will be in one of the following categories: Standard or Enhanced. If not, select None."
    entity = Person
    definition_period = YEAR
    value_type = Enum
    possible_values = PIPCategory
    default_value = PIPCategory.NONE

    def formula(person, period, parameters):
        pip_m = parameters(period).gov.dwp.pip.mobility
        SAFETY_MARGIN = 0.1  # Survey reported values could be slightly below eligible values when they should be above due to data manipulation
        reported_weekly_pip_m = (
            person("PIP_M_reported", period) / WEEKS_IN_YEAR
        )
        return select(
            [
                reported_weekly_pip_m >= pip_m.enhanced * (1 - SAFETY_MARGIN),
                reported_weekly_pip_m >= pip_m.standard * (1 - SAFETY_MARGIN),
                True,
            ],
            [
                PIPCategory.ENHANCED,
                PIPCategory.STANDARD,
                PIPCategory.NONE,
            ],
        )


class pip_dl_category(Variable):
    label = "PIP (daily living) category"
    documentation = "If you receive the daily living component of the Personal Independence Payment, you will be in one of the following categories: Standard or Enhanced. If not, select None."
    entity = Person
    definition_period = YEAR
    value_type = Enum
    possible_values = PIPCategory
    default_value = PIPCategory.NONE

    def formula(person, period, parameters):
        pip_dl = parameters(period).gov.dwp.pip.daily_living
        SAFETY_MARGIN = 0.1  # Survey reported values could be slightly below eligible values when they should be above due to data manipulation
        reported_weekly_pip_dl = (
            person("PIP_DL_reported", period) / WEEKS_IN_YEAR
        )
        return select(
            [
                reported_weekly_pip_dl
                >= pip_dl.enhanced * (1 - SAFETY_MARGIN),
                reported_weekly_pip_dl
                >= pip_dl.standard * (1 - SAFETY_MARGIN),
                True,
            ],
            [
                PIPCategory.ENHANCED,
                PIPCategory.STANDARD,
                PIPCategory.NONE,
            ],
        )
